module FFTPACK
!
!=============================================================================
!      A package of fortran subprograms for the fast fourier
!       transform of periodic and other symmetric sequences
!
!                               by
!
!                       Paul N. Swarztrauber
!     national center for atmospheric research  boulder,colorado 80307
!          which is sponsored by the national science foundation
!
!=============================================================================

! This package consists of programs which perform fast fourier
! transforms for double precision periodic sequences and
! certain other symmetric sequences that are listed below.

! 1.   rffti     initialize  rfftf and rfftb
! 2.   rfftf     forward transform of a real periodic sequence
! 3.   rfftb     backward transform of a real coefficient array

! 4.   ezffti    initialize ezfftf and ezfftb
! 5.   ezfftf    a simplified real periodic forward transform
! 6.   ezfftb    a simplified real periodic backward transform

! 7.  cffti     initialize cfftf and cfftb
! 8.  cfftf     forward transform of a complex periodic sequence
! 9.  cfftb     unnormalized inverse of cfftf

      IMPLICIT NONE

CONTAINS

!=============================================================================
      SUBROUTINE RFFTI (N,WSAVE)
!=============================================================================
      INTEGER :: N
      DOUBLE PRECISION, DIMENSION(2*N+15) :: WSAVE

      IF (N .EQ. 1) RETURN
      CALL RFFTI1 (N,WSAVE(N+1),WSAVE(2*N+1))
      RETURN
      END SUBROUTINE RFFTI

!=============================================================================
      SUBROUTINE RFFTF (N,R,WSAVE)
!=============================================================================
      INTEGER :: N
      DOUBLE PRECISION, DIMENSION(2*N+15) :: WSAVE
      DOUBLE PRECISION, DIMENSION(N) :: R

      IF (N .EQ. 1) RETURN
      CALL RFFTF1 (N,R,WSAVE,WSAVE(N+1),WSAVE(2*N+1))
      RETURN
      END SUBROUTINE RFFTF

!=============================================================================
      SUBROUTINE RFFTB (N,R,WSAVE)
!=============================================================================
      INTEGER :: N
      DOUBLE PRECISION, DIMENSION(2*N+15) :: WSAVE
      DOUBLE PRECISION, DIMENSION(N) :: R

      IF (N .EQ. 1) RETURN
      CALL RFFTB1 (N,R,WSAVE,WSAVE(N+1),WSAVE(2*N+1))
      RETURN
      END SUBROUTINE RFFTB

!=============================================================================
      SUBROUTINE EZFFTI (N,WSAVE)
!=============================================================================
      INTEGER :: N
      DOUBLE PRECISION, DIMENSION(3*N+15) :: WSAVE

      IF (N .EQ. 1) RETURN
      CALL EZFFT1 (N,WSAVE(2*N+1),WSAVE(3*N+1))
      RETURN
      END

!=============================================================================
      SUBROUTINE EZFFTF (N,R,AZERO,A,B,WSAVE)
!=============================================================================
!
!                       VERSION 3  JUNE 1979
!
!=============================================================================
      INTEGER :: N, I, NS2, NS2M
      DOUBLE PRECISION :: AZERO, CF, CFM
      DOUBLE PRECISION, DIMENSION(3*N+15) :: WSAVE
      DOUBLE PRECISION, DIMENSION(N/2) :: A, B
      DOUBLE PRECISION, DIMENSION(N) :: R

      IF (N-2) 101,102,103
  101 AZERO = R(1)
      RETURN
  102 AZERO = .5*(R(1)+R(2))
      A(1) = .5*(R(1)-R(2))
      RETURN
  103 DO 104 I=1,N
         WSAVE(I) = R(I)
  104 CONTINUE
      CALL RFFTF (N,WSAVE,WSAVE(N+1))
      CF = 2./DBLE(N)
      CFM = -CF
      AZERO = .5*CF*WSAVE(1)
      NS2 = (N+1)/2
      NS2M = NS2-1
      DO 105 I=1,NS2M
         A(I) = CF*WSAVE(2*I)
         B(I) = CFM*WSAVE(2*I+1)
  105 CONTINUE
      IF (MOD(N,2) .EQ. 1) RETURN
      A(NS2) = .5*CF*WSAVE(N)
      B(NS2) = 0.
      RETURN
      END


!=============================================================================
      SUBROUTINE EZFFTB (N,R,AZERO,A,B,WSAVE)
!=============================================================================
      INTEGER :: N, NS2, I
      DOUBLE PRECISION :: AZERO
      DOUBLE PRECISION, DIMENSION(3*N+15) :: WSAVE
      DOUBLE PRECISION, DIMENSION(N/2) :: A, B
      DOUBLE PRECISION, DIMENSION(N) :: R

      IF (N-2) 101,102,103
  101 R(1) = AZERO
      RETURN
  102 R(1) = AZERO+A(1)
      R(2) = AZERO-A(1)
      RETURN
  103 NS2 = (N-1)/2
      DO 104 I=1,NS2
         R(2*I) = .5*A(I)
         R(2*I+1) = -.5*B(I)
  104 CONTINUE
      R(1) = AZERO
      IF (MOD(N,2) .EQ. 0) R(N) = A(NS2+1)
      CALL RFFTB (N,R,WSAVE(N+1))
      RETURN
      END


!=============================================================================
      SUBROUTINE CFFTI (N,WSAVE)
!=============================================================================
      INTEGER :: N, IW1, IW2
      DOUBLE PRECISION, DIMENSION(4*N+15) :: WSAVE

      IF (N .EQ. 1) RETURN
      IW1 = N+N+1
      IW2 = IW1+N+N
      CALL CFFTI1 (N,WSAVE(IW1),WSAVE(IW2))
      RETURN
      END


!=============================================================================
      SUBROUTINE CFFTF (N,C,WSAVE)
!=============================================================================
      INTEGER :: N, IW1, IW2
      DOUBLE PRECISION, DIMENSION(4*N+15) :: WSAVE
      COMPLEX*16, DIMENSION(N) :: C

      IF (N .EQ. 1) RETURN
      IW1 = N+N+1
      IW2 = IW1+N+N
      CALL CFFTF1 (N,C,WSAVE,WSAVE(IW1),WSAVE(IW2))
      RETURN
      END

!=============================================================================
      SUBROUTINE CFFTB (N,C,WSAVE)
!=============================================================================
      INTEGER :: N, IW1, IW2
      DOUBLE PRECISION, DIMENSION(4*N+15) :: WSAVE
      COMPLEX*16, DIMENSION(N) :: C

      IF (N .EQ. 1) RETURN
      IW1 = N+N+1
      IW2 = IW1+N+N
      CALL CFFTB1 (N,C,WSAVE,WSAVE(IW1),WSAVE(IW2))
      RETURN
      END


end module FFTPACK
