module ForceAndPressure

!     External module parameters
      use Initiation, only : GridPoints, GammaNew
      use Parameters, only : nTheta, nRad, deltaAng, deltaT, nu, pi, rho, &
                             rCylinder

      implicit none

      private :: PressureForces, FrictionForces, StoreForce, &
                 PressureDistribution, StorePressure, &
                 StoreVorticity
      public ::  Drag, Pressure, SurfaceVorticity

contains ! subroutines

!
!=========================================================================
!                        subroutine Force
!=========================================================================
      subroutine Drag(time)
!=========================================================================
!
!=========================================================================
!     Input
      real(kind=8), intent(in) :: time

!     Local variables
      real(kind=8) :: fdp, fdf
      real(kind=8), dimension(nTheta) :: GammaSurface

!------------------------------ Executable Statements --------------------------
!     Calculate the surface pressure distribution
!      call PressureDistribution(SurfacePressure)

!     Calulate pressure drag
      call PressureForces(fdp)!, SurfacePressure)

!     Calculate the friction drag
      call FrictionForces(fdf, GammaSurface)

!     Write data to a text file
      call StoreForce(11, "Force_Coefficients.txt", time, fdp, fdf)

      return
      end subroutine Drag


!=========================================================================
!                     subroutine PressureForces
!=========================================================================
      subroutine PressureForces(fdp)
!=========================================================================
!
!=========================================================================
!     Input
      real(kind=8), intent(out) :: fdp

!     Local variables
      integer :: j

!------------------------------ Executable Statements --------------------------
!     Calculate the drag coefficient
      fdp = 0.0d0

      do j = 1, nTheta, 1
        fdp = fdp + ( dcos(GridPoints(j,1,3)) * sum(GammaNew(1:j)) )
      end do
      fdp = (deltaAng * fdp * rCylinder * rho)/deltaT

      return
      end subroutine PressureForces


!=========================================================================
!                     subroutine Pressure
!=========================================================================
!      subroutine PressureDistribution(SurfacePressure)
!=========================================================================
!
!=========================================================================
!     External subroutines and variables
!      use HelpFunctions, only : FastIntegrate
!      use Initiation, only : GridPointsOld, GridPointsOldOld
!      use Parameters, only : rOuter, rCylinder

!     Output
!      real(kind=8), dimension(nTheta), intent(out) :: SurfacePressure

!     Local variabels
!      integer :: i, j
!      real(kind=8) :: p, p0, UOld, UNew, UrOld, UrNew, Ut
!      real(kind=8), dimension(nRad) :: fx, x


!------------------------------ Executable Statements --------------------------
!      x = GridPoints(1,:,1)
!      do j = 1, nRad, 1
!       Radial velocity component at t-1
!        UrOld = GridPointsOldOld(1,j,5) * dcos(GridPointsOldOld(1,j,3)) + &
!                GridPointsOldOld(1,j,6) * dsin(GridPointsOldOld(1,j,3))

!       Radial vellocity component at t+1
!        UrNew = GridPoints(1,j,5) * dcos(GridPoints(1,j,3)) + &
!                GridPoints(1,j,6) * dsin(GridPoints(1,j,3))

!       Tangential velocity component at t
!        Ut    = (-1) * GridPointsOld(1,j,5) * dsin(GridPoints(1,j,3)) + &
!                GridPointsOld(1,j,6) * dcos(GridPoints(1,j,3))

!        fx(j) = ((UrNew - UrOld)/(2.0d0*deltaT)) - &
!                (Ut * GridPointsOld(1,j,8))  +  &
!                ( (nu/GridPointsOld(1,j,1)) * &
!                ((GridPointsOld(2,j,8) - &
!                  GridPointsOld(nTheta,j,8))/(2.0d0*deltaAng)) )

!      end do

!     Calulate surface pressure at the 1-th node
!      p = FastIntegrate(x, fx)

!     Calculate p0 -> time derivative of phi at the outer boundary
!      UOld = dsqrt((GridPointsOldOld(1,nRad,5)**2.0d0) + &
!                   (GridPointsOldOld(1,nRad,6)**2.0d0) )
!      UNew = dsqrt((GridPoints(1,nRad,5)**2.0d0) + &
!                   (GridPoints(1,nRad,6)**2.0d0) )

!      if (GridPointsOldOld(1,nRad,5) < 0.0d0 ) UOld = -UOld
!      if (GridPoints(1,nRad,5) < 0.0d0 ) UNew = -UNew

!      p0   = dcos(GridPoints(1,nRad,3))*(rOuter + ((rCylinder**2.0d0)/(rOuter))) * &
!                 ((UNew - UOld)/(2.0d0*deltaT))

!     Surface pressure
!      SurfacePressure(1) = rho * (p - 1.0d0*p0)


!      do i = 2, (nTheta-1), 1
!        do j = 1, nRad, 1
!         Radial velocity component at t-1
!          UrOld = GridPointsOldOld(i,j,5) * dcos(GridPointsOldOld(i,j,3)) + &
!                  GridPointsOldOld(i,j,6) * dsin(GridPointsOldOld(i,j,3))

!         Radial vellocity component at t+1
!          UrNew = GridPoints(i,j,5) * dcos(GridPoints(i,j,3)) + &
!                  GridPoints(i,j,6) * dsin(GridPoints(i,j,3))

!         Tangential velocity component at t
!          Ut    = (-1) * GridPointsOld(i,j,5) * dsin(GridPoints(i,j,3)) + &
!                  GridPointsOld(i,j,6) * dcos(GridPoints(i,j,3))

!          fx(j) = ((UrNew - UrOld)/(2.0d0*deltaT)) - &
!                  (Ut * GridPointsOld(i,j,8))  +  &
!                  ((nu/GridPointsOld(i,j,1)) * &
!                  ((GridPointsOld((i+1),j,8) - &
!                    GridPointsOld((i-1),j,8))/(2.0d0*deltaAng)))

!        end do

!     Calulate surface pressure at the i-th node
!      p = FastIntegrate(x, fx)

!     Calculate p0 -> time derivative of phi at the outer boundary
!      UOld = dsqrt((GridPointsOldOld(i,nRad,5)**2.0d0) + &
!                   (GridPointsOldOld(i,nRad,6)**2.0d0) )
!      UNew = dsqrt((GridPoints(i,nRad,5)**2.0d0) + &
!                   (GridPoints(i,nRad,6)**2.0d0) )

!      if (GridPointsOldOld(i,nRad,5) < 0.0d0 ) UOld = -UOld
!      if (GridPoints(i,nRad,5) < 0.0d0 ) UNew = -UNew

!      p0   = dcos(GridPoints(i,nRad,3))*(rOuter + ((rCylinder**2.0d0)/(rOuter))) * &
!                 ((UNew - UOld)/(2.0d0*deltaT))

!     Surface pressure
!      SurfacePressure(i) =  rho * (p - 1.0d0*p0)

!      end do

!      do j = 1, nRad, 1
!       Radial velocity component at t-1
!        UrOld = GridPointsOldOld(nTheta,j,5) * dcos(GridPointsOldOld(nTheta,j,3)) + &
!                GridPointsOldOld(nTheta,j,6) * dsin(GridPointsOldOld(nTheta,j,3))

!       Radial vellocity component at t+1
!        UrNew = GridPoints(nTheta,j,5) * dcos(GridPoints(nTheta,j,3)) + &
!                GridPoints(nTheta,j,6) * dsin(GridPoints(nTheta,j,3))

!       Tangential velocity component at t
!        Ut    = (-1) * GridPointsOld(nTheta,j,5) * dsin(GridPoints(nTheta,j,3)) + &
!                GridPointsOld(nTheta,j,6) * dcos(GridPoints(nTheta,j,3))

!        fx(j) = ((UrNew - UrOld)/(2.0d0*deltaT)) - &
!                (Ut * GridPointsOld(nTheta,j,8))  +  &
!                ((nu/GridPointsOld(nTheta,j,1)) * &
!                ((GridPointsOld(1,j,8) - &
!                  GridPointsOld((nTheta-1),j,8))/(2.0d0*deltaAng)))

!      end do

!     Calulate surface pressure at the i-th node
!      p = FastIntegrate(x, fx)

!     Calculate p0 -> time derivative of phi at the outer boundary
!      UOld = dsqrt((GridPointsOldOld(nTheta,nRad,5)**2.0d0) + &
!                   (GridPointsOldOld(nTheta,nRad,6)**2.0d0) )
!      UNew = dsqrt((GridPoints(nTheta,nRad,5)**2.0d0) + &
!                   (GridPoints(nTheta,nRad,6)**2.0d0) )

!      if (GridPointsOldOld(nTheta,nRad,5) < 0.0d0 ) UOld = -UOld
!      if (GridPoints(nTheta,nRad,5) < 0.0d0 ) UNew = -UNew

!      p0   = dcos(GridPoints(nTheta,nRad,3))*(rOuter + ((rCylinder**2.0d0)/(rOuter))) * &
!                 ((UNew - UOld)/(2.0d0*deltaT))

!     Surface pressure
!      SurfacePressure(nTheta) =  rho * (p - 1.0d0*p0)

!      return
!      end subroutine PressureDistribution

!=========================================================================
!                        subroutine PressureForces
!=========================================================================
!      subroutine PressureForces(fdp, SurfacePressure)
!=========================================================================
!
!=========================================================================
!     External subroutines
!      use HelpFunctions, only : Integration, FastIntegrate

!     Input
!      real(kind=8), dimension(nTheta), intent(in) :: SurfacePressure

!     Output
!      real(kind=8), intent(out) :: fdp

!     Local variables
!      real(kind=8), dimension(nTheta) :: fx, x

!------------------------------ Executable Statements --------------------------
!     Pressure component in x direction
!      fx = (-1.0d0) * SurfacePressure * dcos(GridPoints(:,1,3))
!      x = GridPoints(:,1,3)
!     Calculate the Pressure force by numerical integration of the pressure
!      fdp = FastIntegrate(x, fx)
!      call Integration(fx, 0.0d0, 2.0d0*pi, fdp) ! Pressure drag

!      return
!      end subroutine PressureForces



!=========================================================================
!                        subroutine FrictionForces
!=========================================================================
      subroutine FrictionForces(fdf, GammaSurface)
!=========================================================================
!
!=========================================================================
!     External subroutines
      use HelpFunctions, only : Integration

!     Output
      real(kind=8), intent(out) :: fdf
      real(kind=8), intent(out), dimension(nTheta) :: GammaSurface

!     Local variables
      real(kind=8), dimension(nTheta) :: fx

!------------------------------ Executable Statements --------------------------
!     Estimate the surface vorticity by Taylor expasion in the radial direction
      GammaSurface = GridPoints(1:nTheta, 2, 8) + &
                   ((GammaNew) / (nu * deltaAng * deltaT * GridPoints(1,1,4)))

!     Calculate the friction force by numerical integration
      fx = (-1) * GammaSurface * sin(GridPoints(:,1,3))
      call Integration(fx, 0.0d0, 2.0d0*pi, fdf) ! Friction drag
      fdf = fdf * nu * rho * rCylinder

      return
      end subroutine FrictionForces

!=========================================================================
!                           StoreForce
!=========================================================================
      subroutine StoreForce(iounit, fileName, time, cdp, cdf)
!=========================================================================
!
!=========================================================================
!     Input
      integer, intent(in) :: iounit
      real(kind=8), intent(in) :: cdp, cdf, time
      character(len=*), intent(in) :: fileName

!     Local variables
      integer :: ios

!--------------------- Write force coefficients to file ------------------------
!     Define output format
8001  format((F10.4),1x)
9001  format((ES20.12),1x)

!     Open file
      open(unit=iounit, file=fileName, iostat=ios, status="old",&
       action="write", position="append")

        if ( ios /= 0 ) then
          print *, 'Error opening file:', fileName
          stop
        end if

!       Dimless time
        write(iounit,8001,iostat=ios, advance='NO') time

!       Cd pressure
        write(iounit,9001,iostat=ios, advance='NO') cdp

!       Cd friction
        write(iounit,9001,iostat=ios, advance='YES') cdf

        if ( ios /= 0 ) then
          print '(A)', "Write error in file ", fileName
          stop
        end if

!     Close file
      close(unit = iounit, iostat=ios, status="keep")

      if ( ios /= 0 ) then
        print *, 'Error closing file:', fileName
        stop
      end if

      return
      end subroutine StoreForce



!=========================================================================
!                           subroutine Pressure
!=========================================================================
      subroutine Pressure(time)
!=========================================================================
!
!=========================================================================
!     Input
      real(kind=8), intent(in) :: time

!     Local variables
      real(kind=8), dimension(nTheta) :: SurfacePressure
      character(len=100) :: fileName

!------------------------------ Executable Statements --------------------------
!     Define output format
4367  format( 'Surface_pressure_at_t*_=_', f10.5'.txt' )
      write(fileName, 4367) time

!     Calulate surface pressure
      call PressureDistribution(SurfacePressure)

!     Store pressure distribution
      call StorePressure(11, trim(fileName), time, SurfacePressure)


      return
      end subroutine Pressure


!=========================================================================
!                           subroutine SurfacePressure
!=========================================================================
      subroutine PressureDistribution(SurfacePressure)
!=========================================================================
!
!=========================================================================
!     Output
      real(kind=8), intent(out), dimension(nTheta) :: SurfacePressure

!     Integer
      integer :: i

!------------------------------ Executable Statements --------------------------
      SurfacePressure = 0.0d0

      do i = 2, nTheta, 1
        SurfacePressure(i) = SurfacePressure(i-1) - ((rho*GammaNew(i)) / deltaT)
      end do

      return
      end subroutine PressureDistribution



!=========================================================================
!                           subroutine StorePressure
!=========================================================================
      subroutine StorePressure(iounit, fileName, time, Pressure)
!=========================================================================
!
!=========================================================================
!     Input
      integer, intent(in) :: iounit
      real(kind=8), intent(in) :: time
      real(kind=8), intent(in), dimension(nTheta) :: Pressure
      character(len=*), intent(in) :: fileName

!     Local variables
      integer :: ios, i

!------------------------------ Executable Statements --------------------------
!     Define output format
8001  format((F10.2),1x)
9001  format((ES20.12),1x)

!     Open file
      open(unit=iounit, file=fileName, iostat=ios, status="replace", &
       action="write", position="rewind")

        if ( ios /= 0 ) then
          print *, 'Error opening file:', fileName
          stop
        end if

!       Title of file
        write(iounit,'(A)',iostat=ios, advance='NO') '# '
        write(iounit,'(A)',iostat=ios, advance='YES') fileName

!       Dimensionless time
        write(iounit,'(A)',iostat=ios, advance='NO') '#  Ut/r =  '
        write(iounit,8001,iostat=ios, advance='YES') time

!       Information about file
        write(iounit,'(A)',iostat=ios, advance='NO') '#  Angular positon     '
        write(iounit,'(A)',iostat=ios, advance='YES') '       Pressure     '


!       Surface pressure
        do i = 1, nTheta, 1
          write(iounit,9001,iostat=ios, advance='NO') (i-1) * deltaAng
          write(iounit,9001,iostat=ios, advance='YES') Pressure(i)
        end do

        write(iounit,9001,iostat=ios, advance='NO') 2 * pi
        write(iounit,9001,iostat=ios, advance='YES') Pressure(1)

!       Add empty lines
        write(iounit,'(A)',iostat=ios, advance='YES') ''
        write(iounit,'(A)',iostat=ios, advance='YES') ''

        if ( ios /= 0 ) then
          print '(A)', "Write error in file ", fileName
          stop
        end if

!     Close file
      close(unit = iounit, iostat=ios, status="keep")

      if ( ios /= 0 ) then
        print *, 'Error closing file:', fileName
        stop
      end if

      return
      end subroutine StorePressure


!=========================================================================
!                           subroutine Pressure
!=========================================================================
      subroutine SurfaceVorticity(time)
!=========================================================================
!
!=========================================================================
!     Input
      real(kind=8), intent(in) :: time

!     Local variables
      real(kind=8) :: cdf
      real(kind=8), dimension(nTheta) :: GammaSurface
      character(len=100) :: fileName


!------------------------------ Executable Statements --------------------------
!     Define output format
4367  format( 'Surface_vorticity_at_t*_=_', f10.5'.txt' )
      write(fileName, 4367) time

!     Calulate surface vorticity
      call FrictionForces(cdf, GammaSurface)

!     Store pressure distribution
      call StoreVorticity(11, trim(fileName), time, GammaSurface)

      return
      end subroutine SurfaceVorticity



!=========================================================================
!                           subroutine StorePressure
!=========================================================================
      subroutine StoreVorticity(iounit, fileName, time, GammaSurface)
!=========================================================================
!
!=========================================================================
!     Input
      integer, intent(in) :: iounit
      real(kind=8), intent(in) :: time
      real(kind=8), intent(in), dimension(nTheta) :: GammaSurface
      character(len=*), intent(in) :: fileName

!     Local variables
      integer :: ios, i

!------------------------------ Executable Statements --------------------------
!     Define output format
8001  format((F10.2),1x)
9001  format((ES20.12),1x)

!     Open file
      open(unit=iounit, file=fileName, iostat=ios, status="replace",&
       action="write", position="rewind")

        if ( ios /= 0 ) then
          print *, 'Error opening file:', fileName
          stop
        end if

!       Title of file
        write(iounit,'(A)',iostat=ios, advance='NO') '# '
        write(iounit,'(A)',iostat=ios, advance='YES') fileName

!       Dimensionless time
        write(iounit,'(A)',iostat=ios, advance='NO') '#  Ut/r =  '
        write(iounit,8001,iostat=ios, advance='YES') time

!       Information about file
        write(iounit,'(A)',iostat=ios, advance='NO') '#  Angular positon     '
        write(iounit,'(A)',iostat=ios, advance='YES') '       Vorticity     '


!       Surface pressure
        do i = 1, nTheta, 1
          write(iounit,9001,iostat=ios, advance='NO') (i-1) * deltaAng
          write(iounit,9001,iostat=ios, advance='YES') GammaSurface(i)
        end do

        write(iounit,9001,iostat=ios, advance='NO') 2 * pi
        write(iounit,9001,iostat=ios, advance='YES') GammaSurface(1)

!       Add empty lines
        write(iounit,'(A)',iostat=ios, advance='YES') ''
        write(iounit,'(A)',iostat=ios, advance='YES') ''

        if ( ios /= 0 ) then
          print '(A)', "Write error in file ", fileName
          stop
        end if

!     Close file
      close(unit = iounit, iostat=ios, status="keep")

      if ( ios /= 0 ) then
        print *, 'Error closing file:', fileName
        stop
      end if

      return
      end subroutine StoreVorticity

!

!=========================================================================
!                           subroutine StorePressure
!=========================================================================
      subroutine StoreVortex(iounit, fileName, time)
!=========================================================================
!
!=========================================================================
!     External variables
      use TypeDefine
      use Initiation, only : Head

!     Input
      integer, intent(in) :: iounit
      real(kind=8), intent(in) :: time
      character(len=*), intent(in) :: fileName

!     Local variables
      integer :: ios
      type(Vortex), pointer :: current
!------------------------------ Executable Statements --------------------------
!     Define output format
8001  format((F10.2),1x)
9001  format((ES20.12),1x)
!     Define output format

!     Open file
      open(unit=iounit, file=fileName, iostat=ios, status="replace", &
       action="write", position="rewind")

        if ( ios /= 0 ) then
          print *, 'Error opening file:', fileName
          stop
        end if

!       Title of file
        write(iounit,'(A)',iostat=ios, advance='NO') '# '
        write(iounit,'(A)',iostat=ios, advance='YES') fileName

!       Dimensionless time
        write(iounit,'(A)',iostat=ios, advance='NO') '#  Ut/D =  '
        write(iounit,8001,iostat=ios, advance='YES') time

!       Information about file
        write(iounit,'(A)',iostat=ios, advance='NO') '#  Strength     '
        write(iounit,'(A)',iostat=ios, advance='NO') '    x     '
        write(iounit,'(A)',iostat=ios, advance='YES') '   y     '



        current => Head
        do
!         Write vortex to textfile
          write(iounit,9001,iostat=ios, advance='NO') current%Gamma
          write(iounit,9001,iostat=ios, advance='NO') current%xCoor
          write(iounit,9001,iostat=ios, advance='YES') current%yCoor

!         Exit loop if last point vortex is reached
          if ( .not. associated(current%Next) ) exit

!         Update pointer
          current => current%Next

       end do

!      Nullify pointer
       nullify(current)

       if ( ios /= 0 ) then
         print '(A)', "Write error in file ", fileName
         stop
       end if

!     Close file
      close(unit = iounit, iostat=ios, status="keep")

      if ( ios /= 0 ) then
        print *, 'Error closing file:', fileName
        stop
      end if

      return
      end subroutine StoreVortex


      end module ForceAndPressure
