module HelpFunctions

      implicit none

      private !none
      public :: intTOstr, Nonzero, cart2pol, Integration, realTOstr, &
                FastIntegrate !, AccurateIntegrate

contains ! 3 subroutines and 1 function


!=========================================================================
!                    Function: intTOstr()
!=========================================================================
      subroutine Nonzero(S, positon, numbersOFnonzero)
!=========================================================================
!     Removes all zeros form the input array S and returns all the
!     nonzero elements of S. The number of elements removed and there
!     postison in S are stored in numbersOFnonzero and positon.
!=========================================================================
!     External modules
      use Parameters, only : tol

!     Input
      real(kind=8), intent(inout), allocatable, dimension(:) :: S

!     Output
      integer, intent(out) :: numbersOFnonzero
      integer, intent(out), allocatable, dimension(:) :: positon

!     Local variabels
      integer :: i, j, Err
      real(kind=8), allocatable, dimension(:) :: temporary

!------------------------------ Executable Statements --------------------------
!     Find number of nonzero elements
      numbersOFnonzero = count( (abs(S) > 0.0d0) , dim=1)

      if (numbersOFnonzero == size(S, dim=1)) then
!       Allocate memory
        allocate(positon(size(S, dim=1)), stat=Err)
        if (Err/= 0) print *, "positon: Allocation request denied"

        positon = (/ (i, i = 1, size(S, dim=1), 1) /)

        return
      else

!       Allocate memory
        allocate(positon(numbersOFnonzero), stat=Err)
        if (Err/= 0) print *, "positon: Allocation request denied"

        allocate(temporary(numbersOFnonzero), stat=Err)
        if (Err /= 0) print *, "temporary: Allocation request denied"

!       Find and store nonzero elements
        j = 0
        do i = 1, size(S, dim=1), 1
          if ( abs(S(i)) > tol ) then
            j = j + 1
            positon(j) = i
            temporary(j) = S(i)
          end if
        end do

!       Resize S
        Err = 0
        if (allocated(S)) deallocate(S, stat=Err)
        if (Err /= 0) print *, "S: Deallocation request denied"

        call move_alloc(temporary,S)

      end if

      return
      end subroutine Nonzero


!
!=========================================================================
!                    subroutine cart2pol
!=========================================================================
      subroutine cart2pol(x, y, r, theta)
!=========================================================================
!     Transform Cartesian to polar coordinates.
!=========================================================================
!     External modules
      use Parameters, only : pi

!     Input
      real(kind=8), intent(in) :: x, y

!     Output
      real(kind=8), intent(out) :: r, theta

!------------------------------ Executable Statements --------------------------
!     Radial coordinate
      r = sqrt( abs(x**2) + abs(y**2) )

!     Azimuthal coordinate
      if ( abs(x) < 1.0d-8 ) then
          if( y < 0 ) then
            theta = -pi/2.0d0
          else
            theta = pi/2.0d0
          end if

      elseif ( x >= 0 ) then ! Quadrant I or Quadrant IV
        theta = datan(y/x)

      elseif ( y > 0 .and. x < 0  ) then ! Quadrant II
        theta = datan(y/x) + pi

      else ! Quadrant III
        theta = datan(y/x) - pi

      end if

      return
      end subroutine cart2pol


!=========================================================================
!                       subroutine Integration
!=========================================================================
      subroutine Integration(fx, a, b, r)
!=========================================================================
!     Preformes numerical integration of the one-dimensional array fx
!     by the use of the trapezoidal rule
!=========================================================================
!     Input
      real(kind=8), intent(in) :: a, b
      real(kind=8), intent(in), dimension(:) :: fx

!     Output
      real(kind=8), intent(out) :: r

!     Local variables
      integer :: n
      real(kind=8) :: h

!------------------------------ Executable Statements --------------------------
      if ( b <=  a) then
        print *, "Error in subroutine Integration: "
        print *, "Lower integration limit (a) was larger then the upper limit "
        print *, "(b). Integration was not preformed and the subroutine "
        print *, "returned zero."
        r = 0.0
        return

      elseif ( size(fx, dim=1) <= 1 ) then
        print *, "Error in subroutine Integration: "
        print *, "Only a singel value f(x) was given. Integration was not "
        print *, "preformed and the subroutine returned zero."
        r = 0.0
        return

      end if

!     Number of intervals (panels)
      n = size(fx, dim=1) - 1

!     Calculate step size
      h = dble((b - a))/dble(n)

!     Preforme numerical integration
      r = (h/2.0d0) * ( fx(1) + fx(n+1) + 2*sum(fx(2:n), dim=1) )

      return
      end subroutine Integration


!=========================================================================
!                       subroutine FastiIntegrate
!=========================================================================
      pure function FastIntegrate(x, y) result(r)
!=========================================================================
!     Calculates the integral of an array y with respect to x using the
!     trapezoid approximation. Note that the mesh spacing of x does not
!     have to be uniform.
!=========================================================================

        real(kind=8), intent(in)  :: x(:)         ! Variable x
        real(kind=8), intent(in)  :: y(size(x))   ! Function y(x)
        real(kind=8)              :: r            ! Integral ∫y(x)·dx

!       Integrate using the trapezoidal rule
        associate(n => size(x))
          r = sum((y(1+1:n-0) + y(1+0:n-1))*(x(1+1:n-0) - x(1+0:n-1)))/2
        end associate
      end function


!=========================================================================
!                    Function: intTOstr
!=========================================================================
       function intTOstr(in) result(out)
!=========================================================================
!     Convert an integer to a string
!=========================================================================

!     --- Local variables ---
      integer, intent(in)  :: in
      character(len=20)  :: out

!------------------------------ Executable Statements --------------------------

      write (out, *) in
      out = adjustl(out)

      return
      end function intTOstr

!=========================================================================
!                    Function: intTOstr
!=========================================================================
       function realTOstr(in) result(out)
!=========================================================================
!     Convert an real of kind 8 to a string
!=========================================================================

!     --- Local variables ---
      real(kind=4), intent(in)  :: in
      character(len=30)  :: out

!------------------------------ Executable Statements --------------------------

      write (out, *) in
      out = adjustl(out)

      return
    end function realTOstr


!=========================================================================
!                    Function: AccurateIntegrate
!=========================================================================
!      function AccurateIntegrate(x, y, a, b) result(r)
!=========================================================================
! This function constructs a piecewise cubic Hermitian interpolation of an array y(x) based on
! discrete numerical data, and subsequently evaluates the integral of the interpolation in the
! range (a,b). Note that the mesh spacing of x does not necessarily have to be uniform.
!=========================================================================
!      real(kind=8), intent(in)  :: x(:)        !! Variable x
!      real(kind=8), intent(in)  :: y(size(x))  !! Function y(x)
!      real(kind=8), intent(in)  :: a           !! Left endpoint
!      real(kind=8), intent(in)  :: b           !! Right endpoint
!      real(kind=8)              :: r           !! Integral ∫y(x)·dx

!      real(kind=8), external    :: dpchqa
!      real(kind=8)              :: d(size(x))
!      integer               :: err

      ! Create a PCHIP interpolation of the input data
!      call dpchez(size(x), x, y, d, .false., 0, 0, err)

      ! Integrate the interpolation in the provided range
!      r = dpchqa(size(x), x, y, d, a, b, err)

!    return
!    end function AccurateIntegrate


end module HelpFunctions
