module Initiation

!     Module Parameters
      use Parameters, only : pi, tol, nRad, nTheta, std, rCylinder, rOuter, &
                             deltaAng, thicknessFactor
      use TypeDefine

      implicit none

      private :: EulerGrid, SystemMatrix, FixedPoint, fun
      public :: rank, info, nVortices, ipiv, Am, Bm, wsave, Absorbed, GammaNew,&
                Head, Tail, UpperDiag, LowerDiag, UpperDiag2, MainDiag, &
                GridPoints, RHS, InitiationOfSystem

      integer :: rank, info, nVortices
      integer, allocatable, dimension(:) :: ipiv
      real(kind=8) :: Am, Bm
      real(kind=8), allocatable, dimension(:) :: wsave, GammaNew, Absorbed
      real(kind=8), allocatable, dimension(:,:) :: UpperDiag, LowerDiag,&
                                                   UpperDiag2, MainDiag
      real(kind=8), allocatable, dimension(:,:,:) :: GridPoints, GridPointsOld,&
                                                     GridPointsOldOld
      complex(kind=8), allocatable, dimension(:,:) :: RHS
      type(Vortex), pointer :: Head
      type(Vortex), pointer :: Tail

contains ! 4 subroutines and 1 functions

!=========================================================================
!                      subroutine: InitiationOfSystem
!=========================================================================
      subroutine InitiationOfSystem
!=========================================================================
!     Initiate the system which will be solved in the VIC
!
!     Returns:
!
!
!
!=========================================================================
!     External subroutines
      use LAPACK, only : DGTTRF, ErrorCheck
      use FFTPACK, only : CFFTI
      use HelpFunctions, only : intTOstr

!     Local variabels
      integer :: n

!     Allocate and assign values to variabels
      rank = nRad - 2
      nVortices = 0
      allocate(UpperDiag(rank-1,(nTheta/2)+1), &
               LowerDiag(rank-1,(nTheta/2)+1), MainDiag(rank,(nTheta/2)+1),&
               UpperDiag2(rank-2,(nTheta/2)+1), GridPoints(nTheta,nRad,8), &
               ipiv(rank), wsave(4*nTheta + 15), RHS(rank,(nTheta/2)+1), &
               GammaNew(nTheta), Absorbed(nTheta), GridPointsOld(nTheta,nRad,8),&
               GridPointsOldOld(nTheta,nRad,8))
      Absorbed = 0.0d0

!     Nullify head and tail poiter
      nullify(Head)
      nullify(Tail)

!------------------------------ Executable Statements --------------------------
!     Calculate Bm by Fixed-Point iterations
      call FixedPoint(fun, 0.01d0, 100)

!     Calculate Am
      Am = log(1 + (std/Bm))

!     Generate the Eularian grid points
      call EulerGrid

!     Create the diagnals of the system matrix
      call SystemMatrix

!     Compute the LU factorization of the real tridiagonal system matrix for
!     modes 0,1,2,3,..., n/2 using elimination with partial pivoting and row
!     interchanges.
      do n = 0, nTheta/2, 1
        call DGTTRF(rank, LowerDiag(:,n+1), MainDiag(:,n+1) ,&
                    UpperDiag(:,n+1), UpperDiag2(:,n+1), ipiv, info)

        call ErrorCheck(info,&
      ' the LU factorization of the real tridiagonal system matrix for '&
       //'mode number '//trim(intTOstr(n)) )

      end do

!     Initializes the array wsave which is used in both  CFFTB and  CFFTF.
!     The prime factorization of n together with a tabulation of the
!     trigonometric functions are computed and stored in wsave.
      call CFFTI(nTheta,wsave)

!     Create output files for the force coefficients, surface pressure and
!     surface vorticity
      call CreateOutputFiles(11, "Force_Coefficients.txt")

      GridPointsOld = GridPoints
      GridPointsOldOld = GridPoints

      return
      end subroutine InitiationOfSystem


!=========================================================================
!                      subroutine: EulerGrid
!=========================================================================
      subroutine EulerGrid
!=========================================================================
!     Generates the eulerian grid
!
!     Returns:
!       GridPoints, dimension(nRad, nTheta, 8)
!       Contains the nodal coordinates in both real and transformed
!       coordinates. Also contains the flow properties such as
!       flow velocity, stream function and vorticity at each
!       node.
!
!       Ex: GridPoints(i,j,:) = [r, i, theta, d, ux, uy , psi , omega]
!
!=========================================================================

!     Local variabels
      integer :: i
      real(kind=8) :: r

!------------------------------ Executable Statements --------------------------

      do i = 0, (nRad-1), 1 ! Loop over all grid rings
        r = Bm*(dexp(Am*dble(i)) - 1.0d0) + rCylinder ! Inverce of the radial transformation
        GridPoints(:,(i+1),1) =  r ! Physical radial coordinat
        GridPoints(:,(i+1),2) =  i ! Transformed radial coordinat
        GridPoints(:,(i+1),4) =  1.0d0/(Am*(Bm + r - rCylinder)) ! d(r)

      end do

      do i = 0, (nTheta-1), 1 ! Loop over all grid rays
        GridPoints((i+1),:,3) = -pi + deltaAng*i  ! Angular coordinat
      end do

      GridPoints(:,:,5:8) = 0.0

      return
      end subroutine EulerGrid


!=========================================================================
!                      subroutine: SystemMatrix
!========================================================================
      subroutine SystemMatrix
!========================================================================
!     Generates the three diagnals UpperDiag (C), MainDiag (B) and
!     LowerDiag (A) for the tridiagonal system matrices
!
!     Example of system matrix:
!       | B1   C1                     |
!       | A1   B2   C2                |
!       |      A2   B3   C3           |
!       |           A3   B4   C4      |
!       |                A4   B5   C5 |
!       |                     A5   B6 |
!
!
!
!     Input:
!       nRad: Number of nodes in radial direction
!       nTheta: Number of nodes in azimuthal direction
!
!     Returns:
!       MainDiag: Array of rank = 1 which containing the nRad - 2
!       elements of the main diagonal ([B1 B2 B3 B4...])
!       UpperDiag: Array of rank = 1 which containing the nRad - 3
!       elements of the upper diagonal ([C1 C2 C3 C4...])
!       LowerDiag: Array of rank = 1 which containing the nRad - 3
!       elements of the lower diagonal ([A1 A2 A3 A4...])
!
!========================================================================

!     Local variabels
      integer :: i, n
      real(kind=8) :: a, b, r

!------------------------------ Executable Statements --------------------------

!     Body boundary conditon (Grid ring number 2)
      r =  GridPoints(1,2,1)
      a = (r**2.0d0)/((Am*(Bm + r - 1.0d0))**2.0d0)
      b = (r)/(Am*(Bm + r - 1.0d0)) - &
          (r**2.0d0)/(Am*(Bm + r - 1.0d0)**2.0d0)

!     Lower, Upper and Main diagonal for the mode 0,1,2,...,n/2
      do n = 0, nTheta/2, 1
        MainDiag(1,(n+1)) = -((n**2.0d0) + (2.0d0*a)) ! B2
        UpperDiag(1,(n+1)) = (a + (b/2.0d0))! C2
      end do

!     Intorier grid points
      do i = 3, (nRad-2), 1 ! Loop over grid rings 3 to Nr-1
        r =  GridPoints(1,i,1)
        a = (r**2.0d0)/((Am*(Bm + r - 1.0d0))**2.0d0)
        b = (r)/(Am*(Bm + r - 1.0d0)) - &
            (r**2.0d0)/(Am*(Bm + r - 1.0d0)**2.0d0)

!       Lower, Upper and Main diagonal for the mode 0,1,2,...,n/2
        do n = 0, nTheta/2, 1
          LowerDiag((i-2),(n+1)) =  (a - (b/2.0d0)) ! Ai
          MainDiag((i-1), (n+1)) = -((n**2.0d0) + (2.0d0*a)) ! Bi
          UpperDiag((i-1),(n+1)) =  (a + (b/2.0d0))! Ci
        end do


      end do

!     A second order backward finit different scheme is used to estimate the
!     Neumann conditon on the the outer boundary (Grid ring Nr-1)
      r =  GridPoints(1,(nRad-1),1)
      a = (r**2.0d0)/((Am*(Bm + r - 1.0d0))**2.0d0)
      b = (r)/(Am*(Bm + r - 1.0d0)) - &
          (r**2.0d0)/(Am*(Bm + r - 1.0d0)**2.0d0)

!     Main for the mode 0,1,2,...,n/2
      do n = 0, nTheta/2, 1
        LowerDiag(nRad-3,(n+1)) = (2.0d0/3.0d0)*(a - b) ! (A(nr-1)
        MainDiag(nRad-2, (n+1)) = (2.0d0/3.0d0)*(b - a) - (n**2.0d0) ! (B(nr-1)
      end do


      return
      end subroutine SystemMatrix


!=================================================================
!                      subroutine: FixedPoint
!=================================================================
      subroutine FixedPoint(f,x0,MaxIter)
!=================================================================
!     Estimate the value of Bm appling fixed point iterations
!     (Assuming that f is an contraction on R see:
!      Banach fixed point theorem)
!
!     Input:
!       f:  the function
!       x0: the initial guess
!       MaxIter: Maximum number of iterations
!
!
!     Returns:
!       The estimate x
!=================================================================

!     Input and Output
      integer, intent(in) :: MaxIter
      real(kind=8), intent(in) :: x0
      real(kind=8), external :: f

!     Local variables
      integer :: k
      real(kind=8) :: fx

!------------------------------ Executable Statements --------------------------
!     Initial guess
      Bm = x0
      do k = 1, MaxIter, 1
!       Evaluate function
        fx = f(Bm)

!       If the solution has converge exit loop
        if (abs(fx - Bm) < tol ) exit

!       Update Bm value
        Bm = fx

      end do

      if (k >= MaxIter) then
        print *, "Error: Solution of Bm did not converge!"
        Stop
      end if

      return
    end subroutine FixedPoint


!=================================================================
!                     Function: fun
!=================================================================
        real(kind=8) function fun(x)
            implicit none
            real(kind=8), intent(in) :: x

            fun = (dlog(1.0d0 + ((thicknessFactor*std)/x)) / &
                   dlog(1.0d0 + ((rOuter-rCylinder)/x)))*(dble(nRad) - 1.0d0)*x

        end function fun


!=========================================================================
!                     Subroutine: CreateForceFile
!=========================================================================
      subroutine CreateOutputFiles(iounit, fileName)
!=========================================================================
!
!=========================================================================
!     Input
      integer, intent(in) :: iounit
      character(len=*), intent(in) :: fileName

!     Local variables
      integer :: ios

!------------------------------ Executable Statements --------------------------
!     Open file
      open(unit=iounit, file=fileName, iostat=ios, status="replace",&
       action="write", position="rewind")

        if ( ios /= 0 ) then
          print *, 'Error opening file:', fileName
          STOP
        end if

!       Title of file
        write(iounit,'(A)',iostat=ios, advance='NO') '# '
        write(iounit,'(A)',iostat=ios, advance='YES') fileName

!       Information about file
        write(iounit,'(A)',iostat=ios, advance='NO') '#       t  '
        write(iounit,'(A)',iostat=ios, advance='NO') '     F pressure      '
        write(iounit,'(A)',iostat=ios, advance='YES') '    F friction      '

        if ( ios /= 0 ) then
          print '(A)', "Write error in file ", fileName
          stop
        end if

!     Close file
      close(unit = iounit, iostat=ios, status="keep")

      if ( ios /= 0 ) then
        print *, 'Error closing file:', fileName
        stop
      end if

      return
      end subroutine CreateOutputFiles

end module Initiation
