module LAPACK
!  -- LAPACK computational routine (version 3.7.0) --
!  -- LAPACK is a software package provided by Univ. of Tennessee,    --
!  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
!     December 2016

      IMPLICIT NONE

CONTAINS
!  =====================================================================
         SUBROUTINE dgttrf( N, DL, D, DU, DU2, IPIV, INFO )
   !
   !     DGTTRF computes an LU factorization of a real tridiagonal matrix A
   !     using elimination with partial pivoting and row interchanges.
   !
   !     The factorization has the form
   !     A = L * U
   !     where L is a product of permutation and unit lower bidiagonal
   !     matrices and U is upper triangular with nonzeros in only the main
   !     diagonal and first two superdiagonals.
   !
   !     .. Scalar Arguments ..
         INTEGER            INFO, N
   !     ..
   !     .. Array Arguments ..
         INTEGER            IPIV( * )
         DOUBLE PRECISION   D( * ), DL( * ), DU( * ), DU2( * )
   !     ..
   !
   !  =====================================================================
   !
   !     .. Parameters ..
         DOUBLE PRECISION   ZERO
         parameter( zero = 0.0d+0 )
   !     ..
   !     .. Local Scalars ..
         INTEGER            I
         DOUBLE PRECISION   FACT, TEMP
   !     ..
   !     .. Intrinsic Functions ..
         INTRINSIC          abs
   !     ..
   !     .. Executable Statements ..
   !
         info = 0
         IF( n.LT.0 ) THEN
            info = -1
            CALL xerbla( 'DGTTRF', -info )
            RETURN
         END IF
   !
   !     Quick return if possible
   !
         IF( n.EQ.0 ) RETURN
   !
   !     Initialize IPIV(i) = i and DU2(I) = 0
   !
         DO 10 i = 1, n
            ipiv( i ) = i
      10 CONTINUE
         DO 20 i = 1, n - 2
            du2( i ) = zero
      20 CONTINUE
   !
         DO 30 i = 1, n - 2
            IF( abs( d( i ) ).GE.abs( dl( i ) ) ) THEN
   !
   !           No row interchange required, eliminate DL(I)
   !
               IF( d( i ).NE.zero ) THEN
                  fact = dl( i ) / d( i )
                  dl( i ) = fact
                  d( i+1 ) = d( i+1 ) - fact*du( i )
               END IF
            ELSE
   !
   !           Interchange rows I and I+1, eliminate DL(I)
   !
               fact = d( i ) / dl( i )
               d( i ) = dl( i )
               dl( i ) = fact
               temp = du( i )
               du( i ) = d( i+1 )
               d( i+1 ) = temp - fact*d( i+1 )
               du2( i ) = du( i+1 )
               du( i+1 ) = -fact*du( i+1 )
               ipiv( i ) = i + 1
            END IF
      30 CONTINUE
         IF( n.GT.1 ) THEN
            i = n - 1
            IF( abs( d( i ) ).GE.abs( dl( i ) ) ) THEN
               IF( d( i ).NE.zero ) THEN
                  fact = dl( i ) / d( i )
                  dl( i ) = fact
                  d( i+1 ) = d( i+1 ) - fact*du( i )
               END IF
            ELSE
               fact = d( i ) / dl( i )
               d( i ) = dl( i )
               dl( i ) = fact
               temp = du( i )
               du( i ) = d( i+1 )
               d( i+1 ) = temp - fact*d( i+1 )
               ipiv( i ) = i + 1
            END IF
         END IF
   !
   !     Check for a zero on the diagonal of U.
   !
         DO 40 i = 1, n
            IF( d( i ).EQ.zero ) THEN
               info = i
               GO TO 50
            END IF
      40 CONTINUE
      50 CONTINUE

         RETURN
       END subroutine dgttrf



!
!  =====================================================================
       SUBROUTINE dgttrs( TRANS, N, NRHS, DL, D, DU, DU2, IPIV, B, LDB,&
                          INFO )
 !
 !     DGTTRS solves one of the systems of equations
 !     A*X = B  or  A**T*X = B,
 !     with a tridiagonal matrix A using the LU factorization computed
 !     by DGTTRF.
 !
 !     .. Scalar Arguments ..
       CHARACTER          TRANS
       INTEGER            INFO, LDB, N, NRHS
 !     ..
 !     .. Array Arguments ..
       INTEGER            IPIV( * )
       DOUBLE PRECISION   B( ldb, * ), D( * ), DL( * ), DU( * ), DU2( * )
 !     ..
 !
 !  =====================================================================
 !
 !     .. Local Scalars ..
       LOGICAL            NOTRAN
       INTEGER            ITRANS, J, JB, NB
 !     ..
 !     .. External Functions ..
      INTEGER            ILAENV
      EXTERNAL           ilaenv
 !     ..
 !     .. Intrinsic Functions ..
       INTRINSIC          max, min
 !     ..
 !     .. Executable Statements ..
 !
       info = 0
       notran = ( trans.EQ.'N' .OR. trans.EQ.'n' )
       IF( .NOT.notran .AND. .NOT.( trans.EQ.'T' .OR. trans.EQ. &
           't' ) .AND. .NOT.( trans.EQ.'C' .OR. trans.EQ.'c' ) ) THEN
          info = -1
       ELSE IF( n.LT.0 ) THEN
          info = -2
       ELSE IF( nrhs.LT.0 ) THEN
          info = -3
       ELSE IF( ldb.LT.max( n, 1 ) ) THEN
          info = -10
       END IF
       IF( info.NE.0 ) THEN
          CALL xerbla( 'DGTTRS', -info )
          RETURN
       END IF
 !
 !     Quick return if possible
 !
       IF( n.EQ.0 .OR. nrhs.EQ.0 )   RETURN
 !
 !     Decode TRANS
 !
       IF( notran ) THEN
          itrans = 0
       ELSE
          itrans = 1
       END IF
 !
 !     Determine the number of right-hand sides to solve at a time.
 !
       IF( nrhs.EQ.1 ) THEN
          nb = 1
       ELSE
          nb = max( 1, ilaenv( 1, 'DGTTRS', trans, n, nrhs, -1, -1 ) )
       END IF
 !
       IF( nb.GE.nrhs ) THEN
          CALL dgtts2( itrans, n, nrhs, dl, d, du, du2, ipiv, b, ldb )
       ELSE
          DO 10 j = 1, nrhs, nb
             jb = min( nrhs-j+1, nb )
             CALL dgtts2( itrans, n, jb, dl, d, du, du2, ipiv, b( 1, j ), &
                          ldb )
    10    CONTINUE
       END IF
 !
 !     End of DGTTRS
 !
       END SUBROUTINE dgttrs

!
!  =====================================================================
       SUBROUTINE dgtts2( ITRANS, N, NRHS, DL, D, DU, DU2, IPIV, B, LDB )
 !
 !     DGTTS2 solves one of the systems of equations
 !     A*X = B  or  A**T*X = B,
 !     with a tridiagonal matrix A using the LU factorization computed
 !     by DGTTRF.
 !
 !     .. Scalar Arguments ..
       INTEGER            ITRANS, LDB, N, NRHS
 !     ..
 !     .. Array Arguments ..
       INTEGER            IPIV( * )
       DOUBLE PRECISION   B( ldb, * ), D( * ), DL( * ), DU( * ), DU2( * )
 !     ..
 !
 !  =====================================================================
 !
 !     .. Local Scalars ..
       INTEGER            I, IP, J
       DOUBLE PRECISION   TEMP
 !     ..
 !     .. Executable Statements ..
 !
 !     Quick return if possible
 !
       IF( n.EQ.0 .OR. nrhs.EQ.0 )  RETURN
 !
       IF( itrans.EQ.0 ) THEN
 !
 !        Solve A*X = B using the LU factorization of A,
 !        overwriting each right hand side vector with its solution.
 !
          IF( nrhs.LE.1 ) THEN
             j = 1
    10       CONTINUE
 !
 !           Solve L*x = b.
 !
             DO 20 i = 1, n - 1
                ip = ipiv( i )
                temp = b( i+1-ip+i, j ) - dl( i )*b( ip, j )
                b( i, j ) = b( ip, j )
                b( i+1, j ) = temp
    20       CONTINUE
 !
 !           Solve U*x = b.
 !
             b( n, j ) = b( n, j ) / d( n )
             IF( n.GT.1 ) &
                b( n-1, j ) = ( b( n-1, j )-du( n-1 )*b( n, j ) ) / &
                                d( n-1 )
             DO 30 i = n - 2, 1, -1
                b( i, j ) = ( b( i, j )-du( i )*b( i+1, j )-du2( i )* &
                              b( i+2, j ) ) / d( i )
    30       CONTINUE
             IF( j.LT.nrhs ) THEN
                j = j + 1
                GO TO 10
             END IF
          ELSE
             DO 60 j = 1, nrhs
 !
 !              Solve L*x = b.
 !
                DO 40 i = 1, n - 1
                   IF( ipiv( i ).EQ.i ) THEN
                      b( i+1, j ) = b( i+1, j ) - dl( i )*b( i, j )
                   ELSE
                      temp = b( i, j )
                      b( i, j ) = b( i+1, j )
                      b( i+1, j ) = temp - dl( i )*b( i, j )
                   END IF
    40          CONTINUE
 !
 !              Solve U*x = b.
 !
                b( n, j ) = b( n, j ) / d( n )
                IF( n.GT.1 ) &
                   b( n-1, j ) = ( b( n-1, j )-du( n-1 )*b( n, j ) ) / &
                                  d( n-1 )
                DO 50 i = n - 2, 1, -1
                   b( i, j ) = ( b( i, j )-du( i )*b( i+1, j )-du2( i )*&
                                 b( i+2, j ) ) / d( i )
    50          CONTINUE
    60       CONTINUE
          END IF
       ELSE
 !
 !        Solve A**T * X = B.
 !
          IF( nrhs.LE.1 ) THEN
 !
 !           Solve U**T*x = b.
 !
             j = 1
    70       CONTINUE
             b( 1, j ) = b( 1, j ) / d( 1 )
             IF( n.GT.1 ) &
                b( 2, j ) = ( b( 2, j )-du( 1 )*b( 1, j ) ) / d( 2 )
             DO 80 i = 3, n
                b( i, j ) = ( b( i, j )-du( i-1 )*b( i-1, j )-du2( i-2 )*&
                              b( i-2, j ) ) / d( i )
    80       CONTINUE
 !
 !           Solve L**T*x = b.
 !
             DO 90 i = n - 1, 1, -1
                ip = ipiv( i )
                temp = b( i, j ) - dl( i )*b( i+1, j )
                b( i, j ) = b( ip, j )
                b( ip, j ) = temp
    90       CONTINUE
             IF( j.LT.nrhs ) THEN
                j = j + 1
                GO TO 70
             END IF
 !
          ELSE
             DO 120 j = 1, nrhs
 !
 !              Solve U**T*x = b.
 !
                b( 1, j ) = b( 1, j ) / d( 1 )
                IF( n.GT.1 ) &
                   b( 2, j ) = ( b( 2, j )-du( 1 )*b( 1, j ) ) / d( 2 )
                DO 100 i = 3, n
                   b( i, j ) = ( b( i, j )-du( i-1 )*b( i-1, j )-&
                               du2( i-2 )*b( i-2, j ) ) / d( i )
   100          CONTINUE
                DO 110 i = n - 1, 1, -1
                   IF( ipiv( i ).EQ.i ) THEN
                      b( i, j ) = b( i, j ) - dl( i )*b( i+1, j )
                   ELSE
                      temp = b( i+1, j )
                      b( i+1, j ) = b( i, j ) - dl( i )*temp
                      b( i, j ) = temp
                   END IF
   110          CONTINUE
   120       CONTINUE
          END IF
       END IF
 !
 !     End of DGTTS2
 !
       END SUBROUTINE dgtts2


!  =====================================================================
       SUBROUTINE zgttrf( N, DL, D, DU, DU2, IPIV, INFO )
 !
 !  -- LAPACK computational routine (version 3.7.0) --
 !  -- LAPACK is a software package provided by Univ. of Tennessee,    --
 !  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
 !     December 2016
 !
 !     .. Scalar Arguments ..
       INTEGER            INFO, N
 !     ..
 !     .. Array Arguments ..
       INTEGER            IPIV( * )
       COMPLEX*16         D( * ), DL( * ), DU( * ), DU2( * )
 !     ..
 !
 !  =====================================================================
 !
 !     .. Parameters ..
       DOUBLE PRECISION   ZERO
       parameter( zero = 0.0d+0 )
 !     ..
 !     .. Local Scalars ..
       INTEGER            I
       COMPLEX*16         FACT, TEMP, ZDUM
 !     ..
 !     .. External Subroutines ..
       EXTERNAL           xerbla
 !     ..
 !     .. Intrinsic Functions ..
       INTRINSIC          abs, dble, dimag
 !     ..
 !     .. Statement Functions ..
       DOUBLE PRECISION   CABS1
 !     ..
 !     .. Statement Function definitions ..
       cabs1( zdum ) = abs( dble( zdum ) ) + abs( dimag( zdum ) )
 !     ..
 !     .. Executable Statements ..
 !
       info = 0
       IF( n.LT.0 ) THEN
          info = -1
          CALL xerbla( 'ZGTTRF', -info )
          RETURN
       END IF
 !
 !     Quick return if possible
 !
       IF( n.EQ.0 )&
          RETURN
 !
 !     Initialize IPIV(i) = i and DU2(i) = 0
 !
       DO 10 i = 1, n
          ipiv( i ) = i
    10 CONTINUE
       DO 20 i = 1, n - 2
          du2( i ) = zero
    20 CONTINUE
 !
       DO 30 i = 1, n - 2
          IF( cabs1( d( i ) ).GE.cabs1( dl( i ) ) ) THEN
 !
 !           No row interchange required, eliminate DL(I)
 !
             IF( cabs1( d( i ) ).NE.zero ) THEN
                fact = dl( i ) / d( i )
                dl( i ) = fact
                d( i+1 ) = d( i+1 ) - fact*du( i )
             END IF
          ELSE
 !
 !           Interchange rows I and I+1, eliminate DL(I)
 !
             fact = d( i ) / dl( i )
             d( i ) = dl( i )
             dl( i ) = fact
             temp = du( i )
             du( i ) = d( i+1 )
             d( i+1 ) = temp - fact*d( i+1 )
             du2( i ) = du( i+1 )
             du( i+1 ) = -fact*du( i+1 )
             ipiv( i ) = i + 1
          END IF
    30 CONTINUE
       IF( n.GT.1 ) THEN
          i = n - 1
          IF( cabs1( d( i ) ).GE.cabs1( dl( i ) ) ) THEN
             IF( cabs1( d( i ) ).NE.zero ) THEN
                fact = dl( i ) / d( i )
                dl( i ) = fact
                d( i+1 ) = d( i+1 ) - fact*du( i )
             END IF
          ELSE
             fact = d( i ) / dl( i )
             d( i ) = dl( i )
             dl( i ) = fact
             temp = du( i )
             du( i ) = d( i+1 )
             d( i+1 ) = temp - fact*d( i+1 )
             ipiv( i ) = i + 1
          END IF
       END IF
 !
 !     Check for a zero on the diagonal of U.
 !
       DO 40 i = 1, n
          IF( cabs1( d( i ) ).EQ.zero ) THEN
             info = i
             GO TO 50
          END IF
    40 CONTINUE
    50 CONTINUE
 !
       RETURN
 !
 !     End of ZGTTRF
 !
       END SUBROUTINE zgttrf

!  =====================================================================
       SUBROUTINE zgttrs( TRANS, N, NRHS, DL, D, DU, DU2, IPIV, B, LDB,&
                          INFO )
  !
  !  -- LAPACK computational routine (version 3.7.0) --
  !  -- LAPACK is a software package provided by Univ. of Tennessee,    --
  !  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
  !     December 2016
  !
  !     .. Scalar Arguments ..
        CHARACTER          TRANS
        INTEGER            INFO, LDB, N, NRHS
  !     ..
  !     .. Array Arguments ..
        INTEGER            IPIV( * )
        COMPLEX*16         B( ldb, * ), D( * ), DL( * ), DU( * ), DU2( * )
  !     ..
  !
  !  =====================================================================
  !
  !     .. Local Scalars ..
        LOGICAL            NOTRAN
        INTEGER            ITRANS, J, JB, NB
  !     ..
  !     .. External Functions ..
        INTEGER            ILAENV
        EXTERNAL           ilaenv
  !     ..
  !     .. External Subroutines ..
        EXTERNAL           xerbla, zgtts2
  !     ..
  !     .. Intrinsic Functions ..
        INTRINSIC          max, min
  !     ..
  !     .. Executable Statements ..
  !
        info = 0
        notran = ( trans.EQ.'N' .OR. trans.EQ.'n' )
        IF( .NOT.notran .AND. .NOT.( trans.EQ.'T' .OR. trans.EQ.&
            't' ) .AND. .NOT.( trans.EQ.'C' .OR. trans.EQ.'c' ) ) THEN
           info = -1
        ELSE IF( n.LT.0 ) THEN
           info = -2
        ELSE IF( nrhs.LT.0 ) THEN
           info = -3
        ELSE IF( ldb.LT.max( n, 1 ) ) THEN
           info = -10
        END IF
        IF( info.NE.0 ) THEN
           CALL xerbla( 'ZGTTRS', -info )
           RETURN
        END IF
  !
  !     Quick return if possible
  !
        IF( n.EQ.0 .OR. nrhs.EQ.0 ) &
           RETURN
  !
  !     Decode TRANS
  !
        IF( notran ) THEN
           itrans = 0
        ELSE IF( trans.EQ.'T' .OR. trans.EQ.'t' ) THEN
           itrans = 1
        ELSE
           itrans = 2
        END IF
  !
  !     Determine the number of right-hand sides to solve at a time.
  !
        IF( nrhs.EQ.1 ) THEN
           nb = 1
        ELSE
           nb = max( 1, ilaenv( 1, 'ZGTTRS', trans, n, nrhs, -1, -1 ) )
        END IF
  !
        IF( nb.GE.nrhs ) THEN
           CALL zgtts2( itrans, n, nrhs, dl, d, du, du2, ipiv, b, ldb )
        ELSE
           DO 10 j = 1, nrhs, nb
              jb = min( nrhs-j+1, nb )
              CALL zgtts2( itrans, n, jb, dl, d, du, du2, ipiv, b( 1, j ),&
                           ldb )
     10    CONTINUE
        END IF
  !
  !     End of ZGTTRS
  !
        END SUBROUTINE zgttrs



!  =====================================================================
       SUBROUTINE xerbla( SRNAME, INFO )
!
!      XERBLA  is an error handler for the LAPACK routines.
!      It is called by an LAPACK routine if an input parameter has an
!      invalid value.  A message is printed and execution stops.
!
!      Installers may consider modifying the STOP statement in order to
!      call system-specific exception-handling facilities.
!     .. Scalar Arguments ..
       CHARACTER*(*)      SRNAME
       INTEGER            INFO
!     ..
!
! =====================================================================

!     .. Intrinsic Functions ..
       INTRINSIC          len_trim
!     ..
!     .. Executable Statements ..

     WRITE( *, fmt = 9999 )srname( 1:len_trim( srname ) ), info

     stop

  9999 FORMAT( ' ** On entry to ', a, ' parameter number ', i2, ' had ', &
            'an illegal value' )

       END SUBROUTINE xerbla



! =====================================================================
      subroutine ErrorCheck(info, ob)
!      ErrorCheck  is an error handler for the LAPACK routines.
!      It is called after an LAPACK routine. If the output parameter
!      has a value unequal to zero a message is printed and execution
!      the program is terminated.
!
! =====================================================================

!     --- Declare variables ---
      implicit none

!     --- Input/Output ---
      CHARACTER(len=*) ob
      integer, intent(inout) :: info

!     --- Check solution ---
      if ( info < 0 ) then
        info = -info
        WRITE( *, fmt = 9999 )info, ob( 1:len_trim( ob ) )
        stop

9999 FORMAT( 'Error: The ', i2,'-th argument of ', a, ' had an illegal value.')

      elseif ( info > 0 ) then
        WRITE( *, fmt = 8888 )info, info, ob( 1:len_trim( ob ) )
        stop

8888 FORMAT( 'Error: U(', i2, ',', i2, ') of ' ,  a, ' is exactly zero.', &
             'The factorization has been completed, but ', &
             'the factor U is exactly singular, and division', &
             'by zero will occur if it is used to solve a', &
             'system of equations.')

      end if
      end subroutine ErrorCheck



end module LAPACK
