module LinkedListOperations

!     Module Parameters
      use TypeDefine
      implicit none

      private ! none
      public :: list_free, list_insert, list_put, list_next, list_remove, &
                list_init, print_list, print_particle, list_count


contains !

!=========================================================================
!                       function list_next
!=========================================================================
      function list_next(self)
!=========================================================================
!     Return the next node after SELF
!=========================================================================

      type(Vortex), pointer :: self
      type(Vortex), pointer :: list_next

!------------------------------ Executable Statements --------------------------

      list_next => self%next

      return
      end function list_next



!=========================================================================
!                   subroutine list_init
!=========================================================================
      subroutine list_init(self, VortexData)
!=========================================================================
!    Initialize the head node SELF and optionally store the provided DATA.
!=========================================================================

      type(Vortex), pointer :: self
      real(kind=8), dimension(3), intent(in), optional :: VortexData

!------------------------------ Executable Statements --------------------------

      allocate(self)
      nullify(self%next)

      if ( present(VortexData) ) then
          self%Gamma = VortexData(1)
          self%xCoor = VortexData(2)
          self%yCoor = VortexData(3)

      end if

      return
      end subroutine list_init


!=========================================================================
!                       function list_next
!=========================================================================
      subroutine list_insert(self, VortexData)
!=========================================================================
!     Insert a list node after SELF containing VortexData
!=========================================================================

      type(Vortex), pointer :: self
      type(Vortex), pointer :: next
      real(kind=8), intent(in), dimension(3) :: VortexData

!------------------------------ Executable Statements --------------------------

!     Create new node
      allocate(next)

!     Add data to the new node
      next%Gamma = VortexData(1)
      next%xCoor = VortexData(2)
      next%yCoor = VortexData(3)
      next%VelX = 0.0d0
      next%VelY = 0.0d0

!     Add the new node to the list after self
      if ( associated(self%next)  ) then
        next%Next => self%next
        self%next => next

      else
        nullify(next%Next)
        self%next => next

      end if

      return
      end subroutine list_insert


!=========================================================================
!                       function list_next
!=========================================================================
      subroutine list_remove(Self, Temp)
!=========================================================================
!     Remove the list node AFTER Self while storing the vortex strength
!     in Temp
!=========================================================================
      type(Vortex), pointer :: self
      type(Vortex), pointer :: nextNode
      real(kind=8), intent(out) :: Temp

!------------------------------ Executable Statements --------------------------
      if ( associated(self%Next) ) then
        nextNode => self%Next

!       Store the nodal data in temp
        Temp = nextNode%Gamma

!       Remove node from list
        if ( associated(nextNode%Next) ) then
          self%Next => nextNode%Next
        else
          nullify(self%Next)
        end if

!       Delete node
        deallocate(nextNode)
        nullify(nextNode)

      end if

      return
      end subroutine list_remove


!=========================================================================
!                       subroutine: list_free
!=========================================================================
      subroutine list_free(self)
!=========================================================================
!     The subroutine traverses the list, deallocating list nodes and
!     their data until the end of the lsit is reached, as indicated by a
!     node with a null next pointer
!=========================================================================

      type(Vortex), pointer :: self
      type(Vortex), pointer :: current
      type(Vortex), pointer :: next

!------------------------------ Executable Statements --------------------------

      current => self
      do while ( associated(current) )
        next => current%next

!       Remove node
        deallocate(current)
        nullify(current)

!       Move to next node in list
        current => next

      end do

      return
      end subroutine list_free


!=========================================================================
!                         subroutine list_put
!=========================================================================
      subroutine list_put(self, VortexData)
!=========================================================================
!     Store the data in list node SELF
!=========================================================================

      type(Vortex), pointer :: self
      real(kind=8), intent(in), dimension(5) :: VortexData

!------------------------------ Executable Statements --------------------------

!     Add data to the new node
      self%Gamma = VortexData(1)
      self%xCoor = VortexData(2)
      self%yCoor = VortexData(3)
      self%VelX  = VortexData(4)
      self%VelY  = VortexData(5)

      return
      end subroutine list_put

!

!=========================================================================
!                         subroutine print_list
!=========================================================================
      subroutine print_particle(self)
!=========================================================================
!     Display the vortex particle SELF
!=========================================================================
!     External subroutines
      use HelpFunctions, only : cart2pol

      type(Vortex), pointer :: self

!     Local variabels
      real(kind=8) :: x, y, r, theta

!------------------------------ Executable Statements --------------------------
!     Converte the positon of the vortex particle to polar coordinates
      x = self%xCoor
      y = self%yCoor
      call cart2pol(x,y,r,theta)

!     Add space
      write(*,"(A)",advance="yes") ""
      write(*,"(A)",advance="yes") ""

!     Labels
      write(*,"(A)",advance="no") "      Gamma   "
      write(*,"(A)",advance="no") "        x      "
      write(*,"(A)",advance="no") "        y       "
      write(*,"(A)",advance="no") "        r     "
      write(*,"(A)",advance="yes")"       theta    "

!     display votex particels
      write(*,"(100g15.5)",advance="no") self%Gamma
      write(*,"(100g15.5)",advance="no") self%xCoor
      write(*,"(100g15.5)",advance="no") self%yCoor
      write(*,"(100g15.5)",advance="no") r
      write(*,"(100g15.5)",advance="yes") theta


!     Add space
      write(*,"(A)",advance="yes") ""
      write(*,"(A)",advance="yes") ""

      end subroutine print_particle


!=========================================================================
!                         subroutine print_list
!=========================================================================
      subroutine print_list(self)
!=========================================================================
!     Display all vortex particle
!=========================================================================
!     External subroutines
      use HelpFunctions, only : cart2pol

      type(Vortex), pointer :: self
      type(Vortex), pointer :: current

!     Local variabels
      real(kind=8) :: x, y, r, theta

!------------------------------ Executable Statements --------------------------
!     Add space
      write(*,"(A)",advance="yes") ""
      write(*,"(A)",advance="yes") ""

!     Labels
      write(*,"(A)",advance="no") "      Gamma   "
      write(*,"(A)",advance="no") "        x      "
      write(*,"(A)",advance="no") "        y       "
      write(*,"(A)",advance="no") "        r     "
      write(*,"(A)",advance="yes")"       theta    "

      current => self
      do while ( associated(current) )
!       Converte the positon of the vortex particle to polar coordinates
        x = current%xCoor
        y = current%yCoor
        call cart2pol(x,y,r,theta)

      !  display votex particels
        write(*,"(100g15.5)",advance="no") current%Gamma
        write(*,"(100g15.5)",advance="no") current%xCoor
        write(*,"(100g15.5)",advance="no") current%yCoor
        write(*,"(100g15.5)",advance="no") r
        write(*,"(100g15.5)",advance="yes") theta

      ! Move to next node in list
        current => current%next

      end do

!     Add space
      write(*,"(A)",advance="yes") ""
      write(*,"(A)",advance="yes") ""

      end subroutine print_list


!=========================================================================
!                       subroutine: list_count
!=========================================================================
      subroutine list_count(self, numberOfNodes)
!=========================================================================
!     Counts the number of nodes after the SELF (also including SELF)
!=========================================================================

      type(Vortex), pointer :: self
      type(Vortex), pointer :: current
      integer :: numberOfNodes
!------------------------------ Executable Statements --------------------------
      numberOfNodes = 0
      current => self
      do while ( associated(current) )
        numberOfNodes = numberOfNodes + 1

!       Move to next node in list
        current => current%next

      end do

      return
      end subroutine list_count

!=========================================================================
!                         function: list_get
!=========================================================================
!      function list_get(self) result(VortexData)
!=========================================================================
!      Return the vortex data stored in the node SELF
!=========================================================================

!      type(Vortex), pointer :: self

!------------------------------ Executable Statements --------------------------

!      integer, dimension(:), pointer :: data
!      data => self%data

!      return
!      end function list_get


end module LinkedListOperations
