module MatrixSolver

!     Module parameters
      use Parameters, only : nTheta, nRad, uTheta, deltaAng, rCylinder
      use Initiation, only : wsave, GridPoints, RHS, UpperDiag, &
                             MainDiag, LowerDiag, UpperDiag2, ipiv, rank, &
                             info, Am, Bm

      implicit none

contains ! 3 subroutines
!=========================================================================
!                      subroutine: Solve
!=========================================================================
      subroutine Solve(timestep)
!=========================================================================
!    Solves the poisson equation
!
!=========================================================================
!     Input
      integer, intent(in) :: timestep

!------------------------------ Executable Statements --------------------------
!     Update RHS of the linear equation system
      call UpdateRHS(timestep)

!     solve the tridiagonal matrices
      call TriDigSolver(timestep)

      end subroutine Solve

!=========================================================================
!                      subroutine: UpdateRHS
!=========================================================================
      subroutine UpdateRHS(timestep)
!=========================================================================
!     Update the right hand side of the linear equation system (Ax = b)
!
!     Returns:
!       RHS:
!
!     Note:
!       1) The fourier coefficients of real perodic sequences
!          (fourier analysis) is calculated with the subroutine cfftf form
!          the dp-version of FFTpack.
!
!       2) The subroutine cttft returns the complex discrete Fourier transform.
!
!       3) The resulting array of Fourier coefficients is conjugate symmetric.
!
!          See : https://docs.oracle.com/cd/E19957-01/806-6147/FFT.html for
!          more information
!
!=========================================================================
!     External subroutines
      use FFTPACK, only : CFFTF

!     Input
      integer, intent(in) :: timestep

!     Local variabels
      integer :: i
      real(kind=8) :: a, b, r
      complex(kind=8), dimension(nTheta) :: BC, omega
      real(kind=8), dimension(size(wsave, dim=1)) :: dummywsave


!------------------------------ Executable Statements --------------------------
!     Calculate the n/2 + 1 first fourier coefficients of the nodal circulation
!     of the interior nodes (mode 0 to nTheta/2 )
      do i = 2, (rank), 1 ! Loop over grid rings 2 to Nr-1

!       Compute the FFT of the nodal vorticity (ŵ) in the azimuthal dir
        dummywsave = wsave
        omega = cmplx(GridPoints(:,i,8), 0. ,kind=8)
        call CFFTF(nTheta,omega,dummywsave)

!       Update RHS for the i-th grid ring (RHS(r,n) = -r²*ŵn)
        RHS(i-1,:) = -(GridPoints(1,i,1)**2) * omega(1:(nTheta/2)+1)

      end do

!     A second order central different scheme is used to estimate the
!     Neumann conditon on the the outer boundary (Grid ring Nr)

!     Compute the FFT of the nodal vorticity (ŵ) in the azimuthal dir
      dummywsave = wsave
      omega = cmplx(GridPoints(:,(rank+1),8), 0. ,kind=8)
      CALL CFFTF(nTheta,omega,dummywsave)

!     Compute the FFT of the free stream velocity (u_theta) at the outer
!     boundary
      dummywsave = wsave
      BC = cmplx(uTheta(:,timestep), 0. ,kind=8)
      CALL CFFTF(nTheta,BC,dummywsave)

!     Update RHS for the Nr-1-th grid ring (RHS(n,r) = -r²*ŵn +
!     (2/3)*(a + b/2)û_theta)
      r =  GridPoints(1,rank+1,1)
      a = (r**2.0d0)/((Am*(Bm + r - 1.0d0))**2.0d0)
      b = (r)/(Am*(Bm + r - 1.0d0)) - &
          (r**2.0d0)/(Am*(Bm + r - 1.0d0)**2.0d0)

      RHS(rank,:) = -(r**2.0d0)*omega(1:(nTheta/2)+1) + &
                     (2.0d0/(3.0d0 * GridPoints(1,nRad,4))) * &
                     (a + (b/2.0d0)) * BC(1:(nTheta/2)+1)

      return
      end subroutine UpdateRHS


!=========================================================================
!                      subroutine: TriDigSolver
!=========================================================================
      subroutine TriDigSolver(timestep)
!=========================================================================
!     Solves the nTheta/2 + 1 linear equation systems and stores the results
!     in the array GridPoints
!
!     Note:
!       1) The tridiagonal matrix systems are solved by the subroutine DGTTRS
!          form LAPACK.
!          DGTTRS solves one of the systems of equations
!          A*X = B  or  A**T*X = B,
!          with a tridiagonal matrix A using the LU factorization computed
!          by DGTTRF.
!
!       2) The real perodic sequence of the stream function is found with the
!          subroutine rfftb from the dp-version of the FFTpack.
!
!       3) The results on the inverse transform are unnormalized and can be
!          normalized by dividing each value by nTheta.
!
!=========================================================================
!     External subroutines and parameters
      use LAPACK, only : DGTTRS, ErrorCheck
      use FFTPACK, only : CFFTB
      use Velocity, only : NodalVelocity
      use HelpFunctions, only : intTOstr

!     Input
      integer, intent(in) :: timestep

!     Local variabels
      integer :: i, n
      real(kind=8), dimension(rank) :: bReal, bImag
      real(kind=8), dimension(size(wsave, dim=1)) :: dummywsave
      complex(kind=8), dimension(nTheta) :: C

!------------------------------ Executable Statements --------------------------
!     Solves the linear equation system for modes 0,1,2,3,..., nTheta/2
      do n = 0, nTheta/2, 1

!       Compute real part of the solution
        bReal = dble(realpart(RHS(:,n+1)))

        call DGTTRS( 'N', rank, 1, LowerDiag(:,n+1), &
                      MainDiag(:,(n+1)), UpperDiag(:,n+1), &
                      UpperDiag2(:,n+1), ipiv, bReal, rank, info)

        call ErrorCheck(info,&
         'the real solution DGTTRS mode number '//trim(intTOstr(n)))

!       Compute Complex part of the solution
        bImag = dble(imagpart(RHS(:,n+1)))

        call DGTTRS( 'N', rank, 1, LowerDiag(:,n+1), &
                      MainDiag(:,(n+1)), UpperDiag(:,n+1), &
                      UpperDiag2(:,n+1), ipiv, bImag, rank, info)

        call ErrorCheck(info,&
        'the imaginary solution of DGTTRS mode number '//trim(intTOstr(n)))

!       Store solution in the complex array RHS
        RHS(:,n+1) = cmplx( bReal, bImag, kind=8)

      end do

!     Preforme a inverse FFT in the azimuthal direction to obtain the nodal
!     values of the stream function
      do i = 1, rank, 1
!       Caculate the fourier coefficients for mode nTheta/2 + 1, nTheta/2 + 2,..
!       by using the symmetry properties of the fourier transformation
        C = (/ RHS(i,1), RHS(i,2:(nTheta/2)+1),&
                  dconjg(RHS(i,(nTheta/2):2:-1)) /)

!       Calculate nodal stream function values
        dummywsave = wsave
        call CFFTB(nTheta, C, dummywsave)
        GridPoints(:,i+1,7) = realpart(C)/nTheta

      end do

!     Calculate nodal flow velocities
      call NodalVelocity(timestep)

      return
      end subroutine TriDigSolver


end module MatrixSolver
