module Random

!     Module Parameters
      implicit none
      double precision, allocatable, dimension(:) :: GuassianX, GuassianY

contains! 3 subroutines

!=========================================================================
!                      subroutine: ReSeed
!=========================================================================
      subroutine ReSeed
!=========================================================================
!      Restarts the state of the pseudorandom number generator used by
!      RANDOM_NUMBER using a seed value extracted from the Fortran
!      subroutine date_and_time.
!
!=========================================================================

!     Local variables
      integer :: k
      integer, dimension(8) :: values
      integer, dimension(:), allocatable :: seed


!------------------------------ Executable Statements --------------------------

!     gfortran subroutine to return date and time information
!     from the real time system clock. Works down to milliseconds
!     and stores the eight return values in array values.
      call date_and_time(VALUES=values)

!     Restart the state of the pseudorandom number generator
!     k = minimum size of seed
      call random_seed(size=k)

!     Allocate memory to seed
      allocate(seed(k))

!     assign information in values to seed
      seed(1:size(values, dim=1)) = values(:)

!     seed the random number generator
      call random_seed(put=seed)

      return
      end subroutine ReSeed

!=========================================================================
!                      subroutine: BoxMuller
!=========================================================================
      subroutine Guassion
!=========================================================================
!     Generate two one dimensional array of length n containing random
!     Guassion distributed numbers with standard diviation std and
!     zero mean.
!
!=========================================================================
!     External modules
      use Parameters, only : lenSequence, std

!     Local variables
      integer :: i
      real(kind=8) :: X1,X2,R,factor
      real(kind=4), dimension(2) :: rVal

!     Allocate memory
      allocate(GuassianX(lenSequence), GuassianY(lenSequence))

!------------------------------ Executable Statements --------------------------

!     Restarts the state of the pseudorandom number generator used by
!     RANDOM_NUMBER
      call ReSeed

!     Generate Guassion distributed variabels using the Box-Mulle scheme
      do i = 1, lenSequence, 1

        call random_number(rVal)
        X1 = 2.*rVal(1) - 1
        X2 = 2.*rVal(2) - 1
        R = X1**2 + X2**2

        do while (R > 1.)
          call random_number(rVal)
          X1 = 2.*rVal(1) - 1
          X2 = 2.*rVal(2) - 1
          R = X1**2 + X2**2
        end do

        factor = sqrt(-2.0*log(R)/R)

        GuassianX(i) = X1 * factor * std
        GuassianY(i) = X2 * factor * std

      end do

      return
      end subroutine Guassion

!=========================================================================
!                      subroutine: Uniform
!=========================================================================
      subroutine Uniform(n, lowerBound, upperBound, UniformSeq)
!=========================================================================
!     Generate a 1-dimensional array of length n containing random
!     numbers uniformly distributed from lowerBound to upperBound
!
!     Input:
!       n: Number of random numbers Generated
!       lowerBound: Lower bound of the uniform distribution
!       upperBound: Upper bound of the uniform distribution
!
!     Returns:
!       UniformSeq: 1 dimensional array of length n containing random
!                   numbers uniformly distributed from lowerBound to
!                   upperBound
!
!=========================================================================

!     Input
      integer, intent(in) :: n, lowerBound, upperBound

!     Output
      integer, intent(out), dimension(n) :: UniformSeq

!     Local variabels
      real(kind=4), dimension(n) :: randomVar

!------------------------------ Executable Statements --------------------------
!     Restarts the state of the pseudorandom number generator used by
!     RANDOM_NUMBER
      call ReSeed

!     Generates a array of pseudorandom numbers from the uniform
!     distribution over the range [0, 1).
      call random_number(randomVar)

!     Generates a array of pseudorandom numbers from the uniform
!     distribution over the range [lowerBound, upperBound].
      UniformSeq = int(randomVar*(upperBound + 1 - lowerBound)) + lowerBound

      return
      end subroutine Uniform

end module Random
