module Velocity

!     Module Parameters
      use Parameters, only : nTheta, nRad, rCylinder
      use Initiation, only : GridPoints, Head
      use TypeDefine

      implicit none

      private ! none
      public :: NodalVelocity, VortexVelocity

contains ! 2 subroutines

!=========================================================================
!                        subroutine: NodalVelocity
!=========================================================================
      subroutine NodalVelocity(timestep)
!=========================================================================
!
!
!
!=========================================================================
!     External subroutines and parameters
      use FFTPACK, only : CFFTB
      use Initiation, only : wsave, RHS
      use Parameters, only : uTheta, uRad

!     Input
      integer, intent(in) :: timestep

!     Local variabels
      integer :: i
      real(kind=8), dimension(nTheta) :: Angle, k
      real(kind=8), dimension(nTheta,(nRad-1)) :: Ut, Ur
      real(kind=8), dimension(size(wsave, dim=1)) :: dummywsave
      complex(kind=8), dimension(nTheta) :: C

!     Assign values to local variabels
      k(1:nTheta) = (/ ( dble(i) , i = 0, (nTheta/2-1), 1 ), 0.0d0, &
                       ( dble(i) , i = (-nTheta/2+1), -1 ) /)

!------------------------------ Executable Statements --------------------------
!     Preformes inverse FFT of n*psi_n (where n is the mode number and psi_n
!     the n-th complex fourier coefficients of psi) in the azimuthal direction
!     to obtain the derivative of the stream funciton in the azimuthal
!     direction and thus the radial velocity component

      Ur(:,1) = 0.0d0
      do i = 1, nRad-2, 1
        C = cmplx(0.0d0, k , kind=8) * (/ RHS(i,1), RHS(i,2:(nTheta/2)+1), &
                                       dconjg(RHS(i,(nTheta/2):2:-1)) /)

        dummywsave = wsave
        call CFFTB(nTheta, C, dummywsave)

        Ur(:,i+1) = ( realpart(C) ) / (nTheta * GridPoints(1,i+1,1) )

      end do

!     A second order forward difference scheme is used to calculate the
!     azimuthal velocity component at the body surface (Grid-ring 1)
      Ut(:,1) = (-0.5d0) * ( GridPoints(1,1,4) * (4.0d0*GridPoints(:,2,7) - &
                                                  GridPoints(:,3,7)))

!     A second order central difference scheme scheme is used to calculate
!     the azimuthal velocity compone at the intorier grid points
!     (Grid-ring 2 to Grid-ring Nr-2)
      do i = 2, (nRad-2), 1
        Ut(:,i) = (-0.5d0) * (GridPoints(1,i,4) * &
                           (GridPoints(:,i+1,7) - GridPoints(:,i-1,7)))
      end do

!     Since psi at the outer boundary is unknown a second order backward
!     difference scheme is employed for Grid-ring Nr-1
      Ut(:,(nRad-1)) = (-0.5d0) * (  GridPoints(1,(nRad-1),4) * &
                                (3.0d0*GridPoints(:,(nRad-1),7) - &
                                 4.0d0*GridPoints(:,(nRad-2),7) + &
                                   GridPoints(:,(nRad-3),7)))

!     Converte nodal velocitys to cartesian components (Grid-rings 1 to Nr-1)
      Angle = GridPoints(:,1,3)
      do i = 1, (nRad-1), 1
        GridPoints(:,i,5) = dcos(Angle)*Ur(:,i) - dsin(Angle)*Ut(:,i) ! Ux
        GridPoints(:,i,6) = dsin(Angle)*Ur(:,i) + dcos(Angle)*Ut(:,i) ! Uy
      end do

!     Converte the velocity components at the boundary to cartesian
!     velocity components
      GridPoints(:,nRad,5) = dcos(Angle)*uRad(:,timestep) - &
                             dsin(Angle)*uTheta(:,timestep) ! Ux
      GridPoints(:,nRad,6) = dsin(Angle)*uRad(:,timestep) + &
                             dcos(Angle)*uTheta(:,timestep) ! Uy

      return
      end subroutine NodalVelocity


!=========================================================================
!                        subroutine: VortexVelocity
!=========================================================================
      subroutine VortexVelocity
!=========================================================================
!
!
!
!=========================================================================
!     External subroutines
      use Vorticity, only :  FindAreaWeigths
      use HelpFunctions, only : cart2pol

!     External parameters
      use Parameters, only : rOuter

!     Local variabels
      integer, dimension(4) :: row, col
      real(kind=8) :: r, theta, x, y
      real(kind=8), dimension(4) :: AreaWeigth
      type(Vortex), pointer :: current


!------------------------------ Executable Statements --------------------------
      current => Head
      do
!       Get vortex positon
        x = current%xCoor
        y = current%yCoor

!       Converte the positon of the vortex from cartesian to polar coordinates
        call cart2pol(x, y, r, theta)

        if ( r >= rCylinder  .and. r <= rOuter ) then ! If vortex particle is within the flow domain
!         Calculate Area weigths
          call FindAreaWeigths(AreaWeigth, row, col, r, theta)

!         Calculate flow velocity at the n-th vortex
!         Ux
          current%VelX = AreaWeigth(1)*GridPoints(row(1),col(1),5) + &
                         AreaWeigth(2)*GridPoints(row(2),col(2),5) + &
                         AreaWeigth(3)*GridPoints(row(3),col(3),5) + &
                         AreaWeigth(4)*GridPoints(row(4),col(4),5)

!         Uy
          current%VelY = AreaWeigth(1)*GridPoints(row(1),col(1),6) + &
                         AreaWeigth(2)*GridPoints(row(2),col(2),6) + &
                         AreaWeigth(3)*GridPoints(row(3),col(3),6) + &
                         AreaWeigth(4)*GridPoints(row(4),col(4),6)

        else
            current%VelX = 0.0d0
            current%VelY = 0.0d0

        end if

!       Exit loop if last point vortex is reached
        if ( .not. associated(current%Next) ) exit

!       Update pointer
        current => current%Next

      end do

!     Nullify pointer
      nullify(current)

      return
      end subroutine VortexVelocity

end module Velocity
