module Vorticity

!     Module Parameters
      use Parameters, only : nRad, nTheta, rOuter, deltaAng, pi, nSub,&
                             tol, nSuba, rCylinder
      use Initiation, only : GridPoints, Am, Bm, Head, Tail, GammaNew
      use TypeDefine

      implicit none

      private :: SubVortices
      public :: NewVoricity, FindAreaWeigths, Projection, Absorb, Reinject, &
                VortexRemove

contains !


!=========================================================================
!                       subroutine: NewVoricity
!=========================================================================
      subroutine NewVoricity
!=========================================================================
!
!
!=========================================================================
!     External subroutines
      use LinkedListOperations, only : list_insert, list_free, list_init

!     Local variabels
      integer :: i, Err
      real(kind=8) :: r, a, b
      real(kind=8), dimension(nTheta) :: omegaNew
      real(kind=8), allocatable, dimension(:,:) :: NewVortices
      type(Vortex), pointer :: current
      nullify(current)

!------------------------------ Executable Statements --------------------------
!     Calculate the nodal vorticity created along the cylinder surface
      r =  rCylinder
      a = (r**2.0d0)/((Am*Bm)**2.0d0)
      b = ((r)/(Am*Bm)) - &
          ((r**2.0d0)/(Am*(Bm)**2.0d0))

      omegaNew = ((-1.0d0) * (a + 0.5d0*b) * GridPoints(:,2,7)) - &
                   GridPoints(:,1,8)

!     Convert voticity to circulaten
      GammaNew = (omegaNew *  r * deltaAng) / ( Gridpoints(1,1,4))

      If ( count( (abs(GammaNew) > 0.0d0 ) , dim=1 ) > 0 ) then
!       Spread the newly generated vorticity unifomly over a segment of the
!       cylinder surface
        call SubVortices(NewVortices)

!       Add the newly created vortices to the end of the list containing the
!       vortex particels

        if ( associated(Head) .and. associated(Tail) ) then

          current => Tail
          do i = 1, size(NewVortices, dim=2), 1
!           Create a new node at the end of the list containing the vortex data
!           for the i-th new vortex
            call list_insert(current, NewVortices(:,i))

!           Move pointer to current last node
            current => current%next

!           Set age of votex
            current%age = 0

          end do

!         Update Tail pointer node
          Tail => current

!         Deallocation memory and nullify local pointers
          Err = 0
          if (allocated(NewVortices)) deallocate(NewVortices, stat=Err)
          if (Err /= 0) print *, "NewVoricity: Deallocation request denied"
          if ( associated(current) ) nullify(current)

        else if ( (.not. associated(Head)) .and. &
                  (.not. associated(Tail)) ) then

!         Initialize the head node
          call list_init(Head, NewVortices(:,1))

          current => Head
          do i = 2, size(NewVortices, dim=2), 1
!           Create a new node at the end of the list containing the vortex data
!           for the i-th new vortex
            call list_insert(current, NewVortices(:,i))

!           Move pointer to current last node
            current => current%next

          end do

!         Update Tail pointer node
          Tail => current

!         Deallocation memory and nullify local pointers
          Err = 0
          if (allocated(NewVortices)) deallocate(NewVortices, stat=Err)
          if (Err /= 0) print *, "NewVoricity: Deallocation request denied"
          if ( associated(current) ) nullify(current)

        else
          print *, "Error in NewVoricity:"
          print *, "Could not access Head/Tail node of vortex list"
          if ( associated(Head) ) call list_free(Head)
          if ( associated(Tail) ) nullify(Tail)
          stop

        end if
      end if

      return
      end subroutine NewVoricity
!
!=========================================================================
!                       subroutine: SubVortices
!=========================================================================
      subroutine SubVortices(NewVortices)
!=========================================================================
!     The vorticity generated at each surface nodes is unifomly spread
!     over a surface segment with width dTheta around the surface nodes.
!      This is done by splitting the nodal vortices into nSub sub-vortices
!
!     Input:
!       GammaNew: The nodal velocitys needed to furfill the no-slip conditon
!       at each node
!
!     Returns:
!       NewVoricity: Contains the strength and positon of the newly
!       created sub-vortices.
!
!       Ex: NewVoricity(i,:) = [Strength, x, y]
!
!=========================================================================
!     External modules
      use HelpFunctions, only : Nonzero
      use Random, only : GuassianX, Uniform
      use Parameters, only : lenSequence

!     Input
      real(kind=8), intent(out), allocatable, dimension(:,:) :: NewVortices

!     Local variabels
      integer :: i, j, n, col, Err, start
      integer, dimension(1) :: ElementNr
      integer, allocatable, dimension(:) :: positon
      real(kind=8) :: r
      real(kind=8), dimension(nSub) :: theta
      real(kind=8), allocatable, dimension(:) :: S, k

!------------------------------ Executable Statements --------------------------
!     Assign values to local variabels
      col = 1
      start = 1

!     Generate a random number from a uniform distribution
      call Uniform(1, 1, lenSequence, ElementNr)

!     Find the surface nodes where new vortices is generated
      S = GammaNew
      call Nonzero(S, positon, n)

!     Allocate memory
      allocate(NewVortices(3,(size(positon, dim=1))*nSub), stat=Err)
      if (Err /= 0) print *, "NewVortices: Allocation request denied"

      if (nSub == 1) then
        allocate(k(nSub))
        k = 0.0d0

      elseif ( mod(nSub,2) == 1  ) then ! nSub odd
        allocate(k(nSub))
        k = (/ (dble(i) , i = -(floor(real(nSub)/2.0)) , &
                               (floor(real(nSub)/2.0)) , 1)  /)

        k = (k*deltaAng)/dble(nSub-1)

      else ! nSub even
        allocate(k(nSub))
        k = (/ (dble(i)/2.0 , i = -(nSub-1) , (nSub-1), 2)  /)

        k = (k*deltaAng)/dble(nSub+2)

      end if

!     Compute positon and strength for each sub-vortex
      if ( positon(1) == 1 ) then
!       Calculate the strength of the sub-vortices
        NewVortices(1,1:nSub) = S(1)/dble(nSub)

!       Spread the sub-vortices over the cylinder surface (theta = -pi)
        theta = k - pi
        theta(1:floor(real(nSub)/2.0)) = theta(1:floor(real(nSub)/2.0)) + &
              2.0d0*pi

            do j = 1, nSub, 1
!             Pick an random radial position form a half guassian distribution
              r = abs(GuassianX(ElementNr(1))) + rCylinder

!             Converte postion to cartesian coordinates
              NewVortices(2,j) = r * dcos(theta(j)) ! X coordinate
              NewVortices(3,j) = r * dsin(theta(j)) ! Y coordinate

!             Pick the next random number from the gaussian distributed array
!             GuassianX
              ElementNr = ElementNr + 1
              if ( ElementNr(1) > lenSequence ) ElementNr = 1

            end do

        col = col + nSub
        start = 2
      end if

      do i = start, size(positon, dim=1), 1
!       Calculate the strength of the sub-vortices
        NewVortices(1,col:col+nSub-1) = S(i)/dble(nSub)

          do j = 0, (nSub-1), 1
!           Pick an random radial position form a half guassian distribution
            r = abs(GuassianX(ElementNr(1))) + rCylinder

!           Converte postion to cartesian coordinates
            NewVortices(2,col:col+j) = r * dcos(k(j+1) + &
                                  GridPoints(positon(i),1,3)) ! X coordinate
            NewVortices(3,col:col+j) = r * dsin(k(j+1) + &
                                  GridPoints(positon(i),1,3)) ! Y coordinate

!           Pick the next random number from the gaussian distributed array
!           GuassianX
            ElementNr = ElementNr + 1
            if ( ElementNr(1) > lenSequence ) ElementNr = 1

          end do

        col = col + nSub
      end do

!     Deallocation memory
      Err = 0
      if (allocated(positon)) deallocate(positon, stat=Err)
      if (Err /= 0) print *, "positon: Deallocation request denied"
      if (allocated(k)) deallocate(k, stat=Err)
      if (Err /= 0) print *, "k: Deallocation request denied"
      if (allocated(S)) deallocate(S, stat=Err)
      if (Err /= 0) print *, "S: Deallocation request denied"


      return
      end subroutine SubVortices



!=========================================================================
!                       subroutine: FindAreaWeigths
!=========================================================================
      subroutine FindAreaWeigths(AreaWeigth, row, col, r, theta)
!=========================================================================
!
!
!
!=========================================================================
!     Input
      real(kind=8), intent(in) :: r, theta

!     Output
      integer, intent(out), dimension(4) :: row, col
      real(kind=8), intent(out), dimension(4) :: AreaWeigth

!     Local variabels
      integer :: jcell, kcell
      real(kind=8) :: cellArea

!------------------------------ Executable Statements --------------------------
!     Find the number of the first node of the cell noted (jcell, kcell)
      jcell = floor( (1.0d0/Am)*dlog( 1.0d0 + ((r-rCylinder)/Bm)) ) + 1 !ring-number
      kcell = floor( (dble(nTheta)/2.0d0)*(1.0d0 + (theta/pi)) ) + 1 !ray-number

      if ( kcell > 0 .and. kcell < nTheta ) then

!       Finds the positon of the coordinates of the corner four nodes of the
!       cell
        row(1) = kcell
        col(1) = jcell

        row(2) = kcell + 1
        col(2) = jcell

        row(3) = kcell + 1
        col(3) = jcell + 1

        row(4) = kcell
        col(4) = jcell + 1

!       Calculate the area of the cell
        cellArea = deltaAng * ( abs(GridPoints(1,col(4),1)**2.0d0 - &
                                    GridPoints(1,col(2),1)**2.0d0) )

!       Calculate the area weigths
        AreaWeigth(1) = ( abs(GridPoints(row(3),col(3),3) - theta) * &
                          abs(GridPoints(row(3),col(3),1)**2.0d0 - r**2.0d0) )/cellArea

        AreaWeigth(2) = ( abs(GridPoints(row(4),col(4),3) - theta) * &
                          abs(GridPoints(row(4),col(4),1)**2.0d0 - r**2.0d0) )/cellArea

        AreaWeigth(3) = ( abs(GridPoints(row(1),col(1),3) - theta) * &
                          abs(GridPoints(row(1),col(1),1)**2.0d0 - r**2.0d0) )/cellArea

        AreaWeigth(4) = ( abs(GridPoints(row(2),col(2),3) - theta) * &
                          abs(GridPoints(row(2),col(2),1)**2.0d0 - r**2.0d0) )/cellArea

      elseif ( kcell == nTheta  ) then
!       Finds the positon of the coordinates of the corner four nodes of the
!       cell
        row(1) = kcell
        col(1) = jcell

        row(2) = 1
        col(2) = jcell

        row(3) = 1
        col(3) = jcell + 1

        row(4) = kcell
        col(4) = jcell + 1

!       Calculate the area of the cell
        cellArea = deltaAng * ( abs(GridPoints(1, col(4), 1)**2.0d0 - &
                                    GridPoints(1, col(2), 1)**2.0d0) )

!       Calculate the area weigths
        AreaWeigth(1) = ( abs(pi - theta) * &
                          abs(GridPoints(row(3),col(3),1)**2.0d0 - r**2.0d0) )/cellArea

        AreaWeigth(2) = ( abs(GridPoints(row(4),col(4),3) - theta) * &
                          abs(GridPoints(row(4),col(4),1)**2.0d0 - r**2.0d0) )/cellArea

        AreaWeigth(3) = ( abs(GridPoints(row(1),col(1),3) - theta) * &
                          abs(GridPoints(row(1),col(1),1)**2.0d0 - r**2.0d0) )/cellArea

        AreaWeigth(4) = ( abs(pi - theta) * &
                          abs(GridPoints(row(2),col(2),1)**2.0d0 - r**2.0d0) )/cellArea
      else
        print *, "Error in subroutine FindAreaWeigths: "
        print *, "Unkown particle postion"
        stop

      end if

      return
      end subroutine FindAreaWeigths



!=========================================================================
!                            subroutine: Projection
!=========================================================================
      subroutine Projection
!=========================================================================
!      Projects the circulaten of each vortex onto the
!      nodes of the polar mesh; this is done by using
!      the so-called Area-Weighting scheme
!=========================================================================
!     External subvortices
      use HelpFunctions, only : cart2pol

!     Local variables
      integer :: i
      integer, dimension(4) :: row, col
      real(kind=8) :: r, theta, x, y
      real(kind=8), dimension(4) :: AreaWeigth
      type(Vortex), pointer :: current

!------------------------------ Executable Statements --------------------------
!     Re-set nodal vorticity
      Gridpoints(:,:,8) = 0.0d0

      current => Head
      do
!       Get vortex positon
        x = current%xCoor
        y = current%yCoor

!       Converte the positon of the vortex from cartesian to polar coordinates
        call cart2pol(x, y, r, theta)

        if ( r >= rCylinder .and. r <= rOuter) then ! If vortex particle is within the flow domain
!         Calculate the four area weigths for the n-th vortex
          call FindAreaWeigths(AreaWeigth, row, col, r, theta)

!         Distribute the vorticity of the n-th vortex to the four
!         closest corner nodes
          do i = 1, 4, 1
            Gridpoints(row(i),col(i),8) = Gridpoints(row(i),col(i),8) + &
                ((AreaWeigth(i) * Gridpoints(row(i),col(i),4)  * &
                  current%Gamma) / ( Gridpoints(row(i), col(i),1) * deltaAng ))

          end do

        end if

!       Exit loop if last point vortex is reached
        if ( .not. associated(current%Next) ) exit

!       Update pointer
        current => current%Next

       end do

!      Nullify pointer
       nullify(current)

       return
       end subroutine Projection


!=========================================================================
!                      subroutine Absorb
!=========================================================================
      subroutine Absorb
!=========================================================================
!
!
!=========================================================================
!     External subroutines
      use LinkedListOperations, only : list_remove
      use HelpFunctions, only : cart2pol
      use Random, only : Uniform

!     External parameters
      use Initiation, only : Head, Tail, Absorbed

!     Local variables
      integer :: NodeNumber
      real(kind=8) :: x, y, r, theta, VortexStrength
      type(Vortex), pointer :: current, Node

!------------------------------ Executable Statements --------------------------
!     Removes the vortices which has enter the cylinder and merge them at the
!     nearest surface node
      current => Head
      do while ( associated(current%Next) )
        Node => current%Next
        x = Node%xCoor
        y = Node%yCoor

!       Converte the position of the vortec to polar coordinates
        call cart2pol(x, y, r, theta)

        if ( r < rCylinder ) then
!         Remove vortex and temporary store the strength
          call list_remove(current, VortexStrength)

!         Find closest surface node
          NodeNumber = nint( (dble(nTheta)/2.0d0)*(1.0d0 + (theta/pi)) ) + 1
          if (NodeNumber > nTheta) NodeNumber = 1

!         Store the vortex
          Absorbed(NodeNumber) = Absorbed(NodeNumber) + VortexStrength

        else
!         Update pointer
          current => current%Next

        end if
      end do

!     Update Tail pointer node
      Tail => current

!     Nullify pointer
      if ( associated(Node) ) nullify(Node)

!     Check Head node
      current => Head
      x = current%xCoor
      y = current%yCoor

!     Converte the position of the vortec to polar coordinates
      call cart2pol(x,y,r,theta)

      if ( r < rCylinder ) then
!       Find closest surface node
        NodeNumber = nint( (dble(nTheta)/2.0d0)*(1.0d0 + (theta/pi)) ) + 1
        if (NodeNumber > nTheta) NodeNumber = 1

!       Store the vortex
        Absorbed(NodeNumber) = Absorbed(NodeNumber) + current%Gamma

!       Update Head pointer
        Head => current%Next

!       Delete node
        deallocate(current)
        nullify(current)

      end if

!     Nullify pointer
      if ( associated(current) ) nullify(current)

      return
      end subroutine Absorb


!=========================================================================
!                      subroutine Reinject
!=========================================================================
      subroutine Reinject
!=========================================================================
!
!
!=========================================================================
!     External subroutines
      use LinkedListOperations, only : list_insert
      use HelpFunctions, only : cart2pol
      use Random, only : Uniform

!     External parameters
      use Initiation, only : Tail, Absorbed
      use Parameters, only : lenSequence
      use Random, only : GuassianX

!     Local variables
      integer :: i, j
      integer, dimension(1) :: ElementNr
      real(kind=8) :: r, theta
      real(kind=8), dimension(3) :: Reinjected
      type(Vortex), pointer :: current

!------------------------------ Executable Statements --------------------------
!     Re-inject the vortices
      If ( count( ( abs(Absorbed) > tol ) , dim=1 ) > 0 ) then

!     Generate a random number from a uniform distribution
      call Uniform(1, 1, lenSequence, ElementNr)

!       Add the absorbed vortices to the end of the list containing the
!       vortex particels
        current => Tail
        do i = 1, size(Absorbed, dim=1), 1
          if ( abs(Absorbed(i)) > tol ) then
!           Create a new node at the end of the list containing the vortex data for
!           the i-th new vortex

            Reinjected(1) = Absorbed(i)/dble(nSuba)
            theta = GridPoints(i,1,3)

            do j = 1, nSuba, 1
!             Pick an random radial position form a half guassian distribution
              r = abs(GuassianX(ElementNr(1))) + rCylinder

!             Converte postion to cartesian coordinates
              Reinjected(2) = r * dcos(theta) ! X coordinate
              Reinjected(3) = r * dsin(theta) ! Y coordinate

!             Add the Reinjected vortex to the end of the list
              call list_insert(current, Reinjected)

!             Pick the next random number from the guassian distributed array
!             GuassianX
              ElementNr = ElementNr + 1
              if ( ElementNr(1) > lenSequence ) ElementNr = 1

            end do
!           Move pointer to current last node
            current => current%next

          end if
        end do

!       Update Tail pointer
        Tail => current

      end if

!     Nullify pointers
      if ( associated(current) ) nullify(current)

!     Remove the reinjected vorticity from the array Absorbed
      Absorbed = 0.0d0

      return
      end subroutine Reinject


!=========================================================================
!                      subroutine VortexRemove
!=========================================================================
      subroutine VortexRemove
!=========================================================================
!
!
!=========================================================================
!     External subroutines
      use LinkedListOperations, only : list_remove
      use HelpFunctions, only : cart2pol

!     External parameters
      use Initiation, only : Head, Tail
      use Parameters, only : rOuter, rCutoff, tCutoff

!     Local variables
      real(kind=8) :: x, y, r, theta, VortexStrength
      type(Vortex), pointer :: current, Node

!------------------------------ Executable Statements --------------------------
!     Removes the vortices which is outside the cut-off radius and older then
!     limit age or outside the outer bounday
      current => Head
      do while ( associated(current%Next) )
        Node => current%Next
        x = Node%xCoor
        y = Node%yCoor

!       Converte the position of the vortec to polar coordinates
        call cart2pol(x, y, r, theta)

        if ( ((r > rCutoff) .and. (Node%age > tCutoff)) .or. (r > rOuter) ) then
!         Remove vortex and temporary store the strength
          call list_remove(current, VortexStrength)

        else
!         Update vortex age
          Node%age = Node%age + 1

!         Update pointer
          current => current%Next
        end if

      end do

!     Update Tail pointer node
      Tail => current

!     Nullify pointer
      if ( associated(Node) ) nullify(Node)

!     Check Head node
      current => Head
      x = current%xCoor
      y = current%yCoor

!     Converte the position of the vortec to polar coordinates
      call cart2pol(x,y,r,theta)

      if ( ((r > rCutoff) .and. (current%age > tCutoff)) .or. (r > rOuter) ) then
!       Update Head pointer
        Head => current%Next

!       Delete node
        deallocate(current)
        nullify(current)

      end if

!     Nullify pointer
      if ( associated(current) ) nullify(current)

      return
      end subroutine VortexRemove


end module Vorticity
