module VorticityTransport

!     Module Parameters
      use Initiation, only : GridPoints, Head
      use TypeDefine
      implicit none

      private ! none
      public :: RandomWalk, Convetion


contains ! 2 subroutines

!=========================================================================
!                      subroutine: RandomWalk
!=========================================================================
      subroutine RandomWalk
!=========================================================================
!
!
!=========================================================================
!     External subroutines and parameters
      use Parameters, only : lenSequence
      use Random, only : Uniform, GuassianX, GuassianY
      use Vorticity, only : Absorb, Reinject
      use HelpFunctions, only : cart2pol

!     Local variables
      integer, dimension(2) :: ElementNr
      type(Vortex), pointer :: current

!------------------------------ Executable Statements --------------------------
!     Generate two random numbers in the interval [1, lenSequence]
      call Uniform(2, 1, lenSequence, ElementNr)

      current => Head
      do
!     Update positon of the n-th vortex
        current%xCoor = current%xCoor + GuassianX(ElementNr(1))
        current%yCoor = current%yCoor + GuassianY(ElementNr(2))

!       Exit loop if last element in list is reached
        if ( .not. associated(current%Next) ) exit

!       Pick the next random number from the guassian distributed arrays
!       GuassianX and GuassianY
        ElementNr(1) = ElementNr(1) + 1
        ElementNr(2) = ElementNr(2) + 1
        if ( ElementNr(1) > lenSequence ) ElementNr(1) = 1
        if ( ElementNr(2) > lenSequence ) ElementNr(2) = 1

!       Update pointer
        current => current%Next

      end do

!     Nullify pointer
      Nullify(current)

!     Vortices that enter the body are removed from the flow and stored in the
!     the array Absorbed
      call Absorb

!     Reinject the vortex particels which has entered the body due to the
!     diffusion process
      call Reinject

      return
      end subroutine RandomWalk
!

!=========================================================================
!                      subroutine: Convetion
!=========================================================================
      subroutine Convetion(timestep)
!=========================================================================
!
!
!=========================================================================
!     External subroutines
      use MatrixSolver, only : Solve
      use Vorticity, only : Projection
      use Velocity, only  : VortexVelocity
      use HelpFunctions, only : cart2pol

!     External parameters
      use Parameters, only : deltaT, rCylinder

!     Input
      integer, intent(in) :: timestep

!     Local variabels
      real(kind=8) :: rReflect, x, y, r, theta
      type(Vortex), pointer :: current

!------------------------------ Executable Statements --------------------------
      current => Head
!     First order guess (Euler scheme)
      do
!       Store current particle position for the n-th particle
        current%OldX = current%xCoor
        current%OldY = current%yCoor

!       First guess of the solution of the convection equation
        current%d1x = current%VelX * deltaT
        current%d1y = current%VelY * deltaT

!       Updata vortex positon
        current%xCoor = current%xCoor + current%d1x
        current%yCoor = current%yCoor + current%d1y

        if ( .not. associated(current%Next) ) exit

!       Shift pointer to next vortex
        current => current%Next

      end do
      
!     Project the vortices onto the grid
      call Projection

!     Solve the Poisson equation and calculate the flow velocity at each node
      call Solve(timestep)

!     Calculate the flow velocities at each vortex particle
      call VortexVelocity

      current => Head
      do
!       Make a second guess on the solution of the convection equation
        current%d2x = current%VelX * deltaT
        current%d2y = current%VelY * deltaT

!       Second order guess (Second order Runge-Kutta)
        current%xCoor = current%OldX + 0.5d0*(current%d1x + current%d2x)
        current%yCoor = current%OldY + 0.5d0*(current%d1y + current%d2y)

        if ( sqrt( ((current%xCoor)**2) + &
                   ((current%yCoor)**2) ) < rCylinder ) then

!           Get vortex positon
            x = current%xCoor
            y = current%yCoor

!           Converte the positon of the vortex from cartesian to polar
!           coordinates
            call cart2pol(x, y, r, theta)

!           Reflect vortex to its mirror image
            rReflect = abs(rCylinder - r) + rCylinder
            current%xCoor = rReflect * cos(theta)
            current%yCoor = rReflect * sin(theta)

        end if

        if ( .not. associated(current%Next) ) exit

!       Shift pointer to next vortex
        current => current%Next

      end do

!     Nullify pointer
      nullify(current)

      return
      end subroutine Convetion


end module VorticityTransport
